#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Attr related class."""

from itertools import chain
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    TypeVar,
    Union,
    overload,
)

from graviti.utility.repr import ReprMixin, ReprType

_T = TypeVar("_T")
_D = TypeVar("_D")


class _AttrDict(Generic[_T], ReprMixin):

    _repr_type = ReprType.MAPPING

    def __init__(self) -> None:
        # An external node is one without child branches,
        # while an internal node has at least one child branch.
        self._internals: Dict[str, _AttrDict[_T]] = {}
        self._externals: Dict[str, _T] = {}

    def __len__(self) -> int:
        return sum(map(len, self._internals)) + len(self._externals)

    def __getitem__(self, key: str) -> _T:
        try:
            return self._getitem(key)
        except KeyError:
            return self.__missing__(key)

    def __setitem__(self, key: str, value: _T) -> None:
        try:
            prefix, suffix = key.split(".", 1)
        except ValueError:
            if key in self._internals:
                raise KeyError("Key prefix conflict") from None

            self._externals.__setitem__(key, value)
            return

        if prefix in self._externals:
            raise KeyError("Key prefix conflict")

        module = self._internals.get(prefix)
        if not module:
            module = _AttrDict()
            self._internals.__setitem__(prefix, module)

        module.__setitem__(suffix, value)

    def __contains__(self, key: object) -> bool:
        if not isinstance(key, str):
            return False

        try:
            self._getitem(key)
        except KeyError:
            return False

        return True

    def __missing__(self, key: str) -> _T:
        raise KeyError(key)

    def __iter__(self) -> Iterator[str]:
        for key, value in self._internals.items():
            for suffix in value:
                yield ".".join((key, suffix))

        yield from self._externals

    def __getattr__(self, key: str) -> Any:
        try:
            return self._externals.__getitem__(key)
        except KeyError:
            pass

        try:
            return self._internals.__getitem__(key)
        except KeyError:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{key}'"
            ) from None

    def __dir__(self) -> Iterable[str]:
        return chain(self._externals, self._internals, super().__dir__())

    def _getitem(self, key: str) -> _T:
        try:
            prefix, suffix = key.split(".", 1)
        except ValueError:
            return self._externals.__getitem__(key)

        return self._internals.__getitem__(prefix).__getitem__(suffix)

    def _repr_head(self) -> str:
        return AttrDict.__name__


class AttrDict(_AttrDict[_T], Mapping[str, _T]):
    """A dict which allows for attr-style access of values."""

    def _repr_head(self) -> str:
        return self.__class__.__name__

    @overload
    def get(self, key: str) -> Optional[_T]:  # pylint: disable=arguments-differ
        ...

    @overload
    def get(self, key: str, default: _D = ...) -> Union[_D, _T]:
        ...

    def get(self, key: str, default: Any = None) -> Any:
        """Return the value for the key if it is in the dict, else default.

        Arguments:
            key: The key for dict, which can be any immutable type.
            default: The value to be returned if key is not in the dict.

        Returns:
            The value for the key if it is in the dict, else default.

        """
        try:
            return self._getitem(key)
        except KeyError:
            return default



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Graviti Python SDK."""

from graviti.__version__ import __version__
from graviti.dataframe import ColumnSeries as Series
from graviti.dataframe import DataFrame
from graviti.file import Audio, File, Image, PointCloud
from graviti.manager import Workspace
from graviti.utility import engine

__all__ = [
    "Audio",
    "DataFrame",
    "File",
    "Image",
    "PointCloud",
    "Series",
    "Workspace",
    "__version__",
    "engine",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the StorageConfig and StorageConfigManager."""

from typing import TYPE_CHECKING, Any, Dict, Generator, Type, TypeVar

from graviti.exception import ResourceNameError
from graviti.manager.common import LIMIT
from graviti.manager.lazy import LazyPagingList
from graviti.openapi import get_storage_config, list_storage_configs, update_storage_configs
from graviti.utility import LazyAttr, ReprMixin, check_type

if TYPE_CHECKING:
    from graviti import Workspace


_T = TypeVar("_T", bound="StorageConfig")


class StorageConfig(ReprMixin):
    """This class defines the structure of the storage config of Graviti Data Platform.

    Arguments:
        workspace: The workspace of the storage config.
        response: The response of the OpenAPI associated with the storage config::

                {
                    "id": <str>
                    "name": <str>
                    "config_type": <str>
                    "backend_type": <str>
                }

    Attributes:
        name: The name of the storage config.
        config_type: The type of the storage config.
        backend_type: The backend type of the storage config.

    """

    _repr_attrs = ("config_type", "backend_type")

    _id = LazyAttr[str]()
    config_type = LazyAttr[str]()
    backend_type = LazyAttr[str]()

    def __init__(self, workspace: "Workspace", name: str) -> None:
        self._workspace = workspace
        self.name = name

    def _init(self, response: Dict[str, Any]) -> None:
        self._id = response["id"]
        self.config_type = response["config_type"]
        self.backend_type = response["backend_type"]

    def _load(self) -> None:
        _workspace = self._workspace
        response = get_storage_config(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self.name,
        )
        self._init(response)

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'

    @classmethod
    def from_response(cls: Type[_T], workspace: "Workspace", response: Dict[str, Any]) -> _T:
        """Create a :class:`StorageConfig` instance from the response.

        Arguments:
            workspace: The workspace of the storage config.
            response: The response of the storage config::

                {
                    "id": <str>
                    "name": <str>
                    "config_type": <str>
                    "backend_type":  <str>
                }

        Returns:
            A :class:`StorageConfig` instance created from the response.

        """
        obj = cls(workspace, response["name"])
        obj._init(response)

        return obj


class StorageConfigManager:
    """This class defines the operations on the storage config in Graviti.

    Arguments:
        workspace: Class :class:`~graviti.workspace.Workspace` instance.

    """

    def __init__(self, workspace: "Workspace") -> None:
        self._workspace = workspace

    def _generate(self, offset: int, limit: int) -> Generator[StorageConfig, None, int]:
        _workspace = self._workspace
        response = list_storage_configs(
            _workspace.access_key, _workspace.url, _workspace.name, limit=limit, offset=offset
        )

        for item in response["storage_configs"]:
            yield StorageConfig.from_response(_workspace, item)

        return response["total_count"]  # type: ignore[no-any-return]

    @property
    def default_storage_config(self) -> str:
        """The default storage config of this worksapce.

        Returns:
            The default storage config of this worksapce.

        """
        _workspace = self._workspace
        response = list_storage_configs(
            _workspace.access_key, _workspace.url, _workspace.name, limit=1, offset=0
        )
        return response["default_storage_config"]  # type: ignore[no-any-return]

    def get(self, name: str) -> StorageConfig:
        """Get a Graviti storage config with given name.

        Arguments:
            name: The name of the storage config.

        Returns:
            The requested :class:`~graviti.manager.storage_config.storage_config` instance.

        Raises:
            ResourceNameError: When the given storage config name is invalid.

        """
        check_type("name", name, str)
        if not name:
            raise ResourceNameError("storage config", name)

        _workspace = self._workspace
        response = get_storage_config(
            _workspace.access_key, _workspace.url, _workspace.name, storage_config=name
        )

        return StorageConfig.from_response(_workspace, response)

    def list(self) -> LazyPagingList[StorageConfig]:
        """List Graviti storage configs.

        Returns:
            The LazyPagingList of :class:`~graviti.manager.storage_config.StorageConfig` instances.

        """
        return LazyPagingList(self._generate, LIMIT)

    def edit(self, *, default_storage_config: str) -> None:
        """Update Graviti storage config related configs.

        Arguments:
            default_storage_config: The name of the new default storage config.

        Raises:
            ResourceNameError: When the given default storage config name is invalid.

        """
        check_type("default_storage_config", default_storage_config, str)
        if not default_storage_config:
            raise ResourceNameError("storage config", default_storage_config)

        _workspace = self._workspace
        update_storage_configs(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            default_storage_config=default_storage_config,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Definitions of different operations on a DataFrame."""

from typing import TYPE_CHECKING, Dict, Iterable, List, Tuple

from tqdm import tqdm

from graviti.exception import ObjectCopyError
from graviti.file import File, FileBase, RemoteFile
from graviti.openapi import (
    RECORD_KEY,
    add_data,
    copy_objects,
    delete_data,
    update_data,
    update_schema,
)
from graviti.operation.common import get_schema
from graviti.portex import record
from graviti.utility import chunked, submit_multithread_tasks

if TYPE_CHECKING:
    from graviti.dataframe import DataFrame
    from graviti.dataframe.column.series import NumberSeries
    from graviti.manager import Dataset, ObjectPermissionManager

_MAX_BATCH_SIZE = 2048
_MAX_ITEMS = 60000


class DataFrameOperation:
    """This class defines the basic method of the operation on a DataFrame."""

    def get_file_count(self) -> int:  # pylint: disable=no-self-use
        """Get the file amount to be uploaded.

        Returns:
            The file amount to be uploaded.

        """
        return 0

    def get_data_count(self) -> int:  # pylint: disable=no-self-use
        """Get the data amount to be uploaded.

        Returns:
            The data amount to be uploaded.

        """
        return 0

    def execute(
        self,
        dataset: "Dataset",
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
    ) -> None:
        """Execute the OpenAPI create sheet.

        Arguments:
            dataset: The Dataset instance.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class DataOperation(DataFrameOperation):  # pylint: disable=abstract-method
    """This class defines the basic method of the data operation on a DataFrame."""

    def __init__(self, data: "DataFrame") -> None:
        self._data = data

    def _get_max_batch_size(self) -> int:
        return min(
            _MAX_BATCH_SIZE,
            _MAX_ITEMS // self._data.schema._get_column_count(),  # pylint: disable=protected-access
        )

    def get_file_count(self) -> int:
        """Get the file amount to be uploaded.

        Returns:
            The file amount to be uploaded.

        """
        return sum(map(len, self._data._generate_file_series()))  # pylint: disable=protected-access

    def get_data_count(self) -> int:
        """Get the data amount to be uploaded.

        Returns:
            The data amount to be uploaded.

        """
        return len(self._data)


class AddData(DataOperation):
    """This class defines the operation that add data to a DataFrame.

    Arguments:
        data: The data to be added.

    """

    def execute(
        self,
        dataset: "Dataset",
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
    ) -> None:
        """Execute the OpenAPI add data.

        Arguments:
            dataset: The Dataset instance.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.

        """
        batch_size = self._get_max_batch_size()
        df = self._data

        for i in range(0, len(df), batch_size):
            batch = df.iloc[i : i + batch_size]

            # pylint: disable=protected-access
            local_files, remote_files = _separate_files(batch._generate_file(), dataset)
            if remote_files:
                _copy_files(dataset, remote_files, file_pbar)

            if local_files:
                _upload_files(local_files, dataset.object_permission_manager, file_pbar, jobs)

            _workspace = dataset.workspace
            add_data(
                _workspace.access_key,
                _workspace.url,
                _workspace.name,
                dataset.name,
                draft_number=draft_number,
                sheet=sheet,
                data=batch.to_pylist(_to_backend=True),
            )
            data_pbar.update(len(batch))


class UpdateSchema(DataFrameOperation):
    """This class defines the operation that update the schema of a DataFrame.

    Arguments:
        schema: New portex schema after updated.

    """

    def __init__(self, schema: record, data: "DataFrame") -> None:
        self._data = data
        self.schema = schema

    def execute(  # pylint: disable=unused-argument
        self,
        dataset: "Dataset",
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
    ) -> None:
        """Execute the OpenAPI update schema.

        Arguments:
            dataset: The Dataset instance.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.

        """
        portex_schema, avro_schema, arrow_schema = get_schema(self.schema)

        # Request data between "Update Schema" and "Update Data" will make Graviti backend report
        # error. So here to move the "request data" before "UpdateSchema" to bypass this issue.
        record_keys: "NumberSeries" = self._data[RECORD_KEY]  # type: ignore[assignment]
        for page in record_keys._data._pages:  # pylint: disable=protected-access
            page.get_array()

        _workspace = dataset.workspace
        update_schema(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            dataset.name,
            draft_number=draft_number,
            sheet=sheet,
            _schema=portex_schema,
            _avro_schema=avro_schema,
            _arrow_schema=arrow_schema,
        )


class UpdateData(DataOperation):
    """This class defines the operation that updates the data of a DataFrame.

    Arguments:
        data: The data for updating.

    """

    def execute(
        self,
        dataset: "Dataset",
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
    ) -> None:
        """Execute the OpenAPI add data.

        Arguments:
            dataset: The Dataset instance.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.

        """
        batch_size = self._get_max_batch_size()
        df = self._data

        for i in range(0, len(df), batch_size):
            batch = df.iloc[i : i + batch_size]

            # pylint: disable=protected-access
            local_files, remote_files = _separate_files(batch._generate_file(), dataset)
            if remote_files:
                _copy_files(dataset, remote_files, file_pbar)

            if local_files:
                _upload_files(local_files, dataset.object_permission_manager, file_pbar, jobs)

            _workspace = dataset.workspace
            update_data(
                _workspace.access_key,
                _workspace.url,
                _workspace.name,
                dataset.name,
                draft_number=draft_number,
                sheet=sheet,
                data=batch.to_pylist(_to_backend=True),
            )
            data_pbar.update(len(batch))


class DeleteData(DataFrameOperation):
    """This class defines the operation that delete the data of a DataFrame.

    Arguments:
        record_keys: The record keys of the data to be deleted.

    """

    def __init__(self, record_keys: List[str]) -> None:
        self.record_keys = record_keys

    def execute(  # pylint: disable=unused-argument
        self,
        dataset: "Dataset",
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
    ) -> None:
        """Execute the OpenAPI delete data.

        Arguments:
            dataset: The Dataset instance.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.

        """
        _workspace = dataset.workspace
        delete_data(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            dataset.name,
            draft_number=draft_number,
            sheet=sheet,
            record_keys=self.record_keys,
        )


def _copy_files(
    dataset: "Dataset",
    remote_files: Dict[str, List[RemoteFile]],
    pbar: tqdm,
) -> None:
    _workspace = dataset.workspace
    for source_dataset, files in remote_files.items():
        for batch in chunked(files, _MAX_BATCH_SIZE):
            keys = copy_objects(
                _workspace.access_key,
                _workspace.url,
                _workspace.name,
                dataset.name,
                source_dataset=source_dataset,
                keys=[file.key for file in batch],
            )["keys"]
            for file, key in zip(batch, keys):
                file._post_key = key  # pylint: disable=protected-access

            pbar.update(len(batch))


def _upload_files(
    files: Iterable[File],
    object_permission_manager: "ObjectPermissionManager",
    pbar: tqdm,
    jobs: int = 8,
) -> None:
    submit_multithread_tasks(
        lambda file: _upload_file(file, object_permission_manager, pbar),
        files,
        jobs=jobs,
    )


def _upload_file(
    file: File,
    object_permission_manager: "ObjectPermissionManager",
    pbar: tqdm,
) -> None:
    post_key = f"{object_permission_manager.prefix}{file.get_checksum()}"
    object_permission_manager.put_object(post_key, file.path)
    file._post_key = post_key  # pylint: disable=protected-access
    pbar.update()


def _separate_files(
    files: Iterable[FileBase], target_dataset: "Dataset"
) -> Tuple[List[File], Dict[str, List[RemoteFile]]]:
    local_files: List[File] = []
    remote_files: Dict[str, List[RemoteFile]] = {}

    for file in files:
        # pylint: disable=protected-access
        if isinstance(file, File):
            local_files.append(file)

        elif isinstance(file, RemoteFile):
            source_dataset = file._object_permission._dataset
            if source_dataset._id == target_dataset._id:
                continue

            try:
                remote_files[source_dataset.name].append(file)
            except KeyError:
                if source_dataset.workspace._id != target_dataset.workspace._id:
                    raise ObjectCopyError(
                        "It is not allowed to copy object between diffenent workspaces.\n"
                        "  Source:\n"
                        f'    workspace: "{source_dataset.workspace.name}"\n'
                        f'    dataset: "{source_dataset.name}"\n'
                        f'    object key: "{file.key}"\n'
                        "  Target:\n"
                        f'    workspace: "{target_dataset.workspace.name}"\n'
                        f'    dataset: "{target_dataset.name}"\n'
                    ) from None

                if source_dataset.storage_config.name != target_dataset.storage_config.name:
                    raise ObjectCopyError(
                        "It is not allowed to copy object between datasets "
                        "with different storage configs.\n"
                        "  Source:\n"
                        f'    workspace: "{source_dataset.workspace.name}"\n'
                        f'    dataset: "{source_dataset.name}"\n'
                        f'    storage config: "{source_dataset.storage_config.name}"\n'
                        f'    object key: "{file.key}"\n'
                        "  Target:\n"
                        f'    workspace: "{target_dataset.workspace.name}"\n'
                        f'    dataset: "{target_dataset.name}"\n'
                        f'    storage config: "{target_dataset.storage_config.name}"\n'
                    ) from None

                remote_files[source_dataset.name] = [file]

        else:
            raise TypeError("The file instance is neither 'File' nor 'RemoteFile'")

    return local_files, remote_files



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Page related class."""

from typing import Any, Callable, Iterator, Optional, Sequence, TypeVar, Union, overload

_T = TypeVar("_T")


class PageBase(Sequence[_T]):
    """PageBase is the base class of array wrapper and represents a page in paging list."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__len__()})"

    def __len__(self) -> int:
        raise NotImplementedError

    def __iter__(self) -> Iterator[_T]:
        return self._iter()

    @overload
    def __getitem__(self, index: int) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> "PageBase[_T]":
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[_T, "PageBase[_T]"]:
        if isinstance(index, slice):
            return self.get_slice(index.start, index.stop, index.step)

        return self.get_item(index)

    def _patch(self, array: Sequence[_T]) -> None:
        # https://github.com/python/mypy/issues/708
        # https://github.com/python/mypy/issues/2427
        self._iter = array.__iter__  # type: ignore[assignment]
        self.get_item = array.__getitem__  # type: ignore[assignment]

    def _iter(self) -> Iterator[_T]:  # pylint: disable=method-hidden
        return self.get_array().__iter__()

    def get_item(self, index: int) -> _T:  # pylint: disable=method-hidden
        """Return the item at the given index.

        Arguments:
            index: Position of the mutable sequence.

        Returns:
            The item at the given index.

        """
        return self.get_array().__getitem__(index)

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "PageBase[_T]":
        """Return a sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def get_array(self) -> Sequence[_T]:
        """Get the array inside the page.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class Page(PageBase[_T]):
    """Page is an array wrapper and represents a page in paging list.

    Arguments:
        array: The internal sequence of page.

    """

    def __init__(self, array: Sequence[_T]):
        self._array = array
        self._patch(array)

    def __len__(self) -> int:
        return len(self._array)

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "SlicedPage[_T]":
        """Return a sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        return SlicedPage(range(len(self._array))[start:stop:step], self._array)

    def get_array(self) -> Sequence[_T]:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        return self._array


class SlicedPage(PageBase[_T]):
    """SlicedPage is an array wrapper and represents a sliced page in paging list.

    Arguments:
        ranging: The range instance of this page.
        array: The internal sequence of page.

    """

    _array: Optional[Sequence[_T]] = None

    def __init__(self, ranging: range, source_array: Sequence[_T]) -> None:
        self._ranging = ranging
        self._source_array = source_array

    def __len__(self) -> int:
        return len(self._ranging)

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "SlicedPage[_T]":
        """Return a sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        return SlicedPage(self._ranging[start:stop:step], self._source_array)

    def get_array(self) -> Sequence[_T]:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        array = self._array
        if array is None:
            ranging = self._ranging
            stop = ranging.stop
            array = self._source_array[ranging.start : stop if stop != -1 else None : ranging.step]
            self._array = array
            self._patch(array)

        return array


class LazyPage(PageBase[_T]):
    """LazyPage is a placeholder when the paging list page is not loaded yet.

    Arguments:
        length: The length of this page.
        array_getter: A callable object to get the source array.

    """

    _array: Optional[Sequence[_T]] = None

    def __init__(self, length: int, array_getter: Callable[[], Sequence[_T]]) -> None:
        self._length = length
        self._array_getter = array_getter

    def __len__(self) -> int:
        return self._length

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "LazySlicedPage[_T]":
        """Return a lazy sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        return LazySlicedPage(range(self._length)[start:stop:step], self.get_array)

    def get_array(self) -> Sequence[_T]:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        array = self._array
        if array is None:
            array = self._array_getter()
            self._array = array
            self._patch(array)

        return array


class LazySlicedPage(PageBase[_T]):
    """LazySlicedPage is a placeholder when the sliced paging list page is not loaded yet.

    Arguments:
        ranging: The range instance of this page.
        array_getter: A callable object to get the source array.

    """

    _array: Optional[Sequence[_T]] = None

    def __init__(self, ranging: range, array_getter: Callable[[], Sequence[_T]]) -> None:
        self._ranging = ranging
        self._array_getter = array_getter

    def __len__(self) -> int:
        return len(self._ranging)

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "LazySlicedPage[_T]":
        """Return a lazy sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        return LazySlicedPage(self._ranging[start:stop:step], self._array_getter)

    def get_array(self) -> Sequence[_T]:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        array = self._array
        if array is None:
            ranging = self._ranging
            stop = ranging.stop
            array = self._array_getter()[
                ranging.start : stop if stop != -1 else None : ranging.step
            ]

            self._array = array
            self._patch(array)

        return array


class MappedPageBase(PageBase[_T]):
    """MappedPageBase is the base class of the page with mapper, is used for nested DataFrame."""

    def copy(
        self, copier: Callable[[Any], Any], mapper: Callable[[Any], Any]
    ) -> "MappedPageBase[_T]":
        """Return a copy of the mapped page.

        Arguments:
            copier: A callable object to convert loaded items in the source page to the copied page.
            mapper: The mapper of the new mapped page.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class MappedPage(MappedPageBase[_T], Page[_T]):
    """MappedPage is an array wrapper and represents a page in paging list."""

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "MappedSlicedPage[_T]":
        """Return a sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        return MappedSlicedPage(range(len(self._array))[start:stop:step], self._array)

    def copy(self, copier: Callable[[_T], _T], mapper: Callable[[Any], Any]) -> "MappedPage[_T]":
        """Return a copy of the mapped page.

        Arguments:
            copier: A callable object to convert loaded items in the source page to the copied page.
            mapper: The mapper of the new mapped page.

        Returns:
            A copy of the mapped page.

        """
        return MappedPage(tuple(map(copier, self._array)))


class MappedSlicedPage(MappedPageBase[_T], SlicedPage[_T]):
    """MappedSlicedPage is an array wrapper and represents a sliced page in paging list."""

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "MappedSlicedPage[_T]":
        """Return a sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        return MappedSlicedPage(self._ranging[start:stop:step], self._source_array)

    def copy(self, copier: Callable[[Any], Any], mapper: Callable[[Any], Any]) -> MappedPage[_T]:
        """Return a copy of the mapped page.

        Arguments:
            copier: A callable object to convert loaded items in the source page to the copied page.
            mapper: The mapper of the new mapped page.

        Returns:
            A copy of the mapped page.

        """
        return MappedPage(tuple(map(copier, self.get_array())))


class MappedLazyPage(MappedPageBase[_T]):
    """LazyPage with a mapper for converting every item in the source array.

    Arguments:
        length: The length of this page.
        array_getter: A callable object to get the source array.
        mapper: A callable object to convert every item in the source array.

    """

    _array: Optional[Sequence[_T]] = None

    def __init__(
        self,
        length: int,
        array_getter: Callable[[], Sequence[_T]],
        mapper: Callable[[Any], Any],
    ) -> None:
        self._length = length
        self._array_getter = array_getter
        self._mapper = mapper

    def __len__(self) -> int:
        return self._length

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "Union[MappedLazySlicedPage[_T], MappedSlicedPage[_T]]":
        """Return a lazy sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        if self._array is not None:
            return MappedSlicedPage(range(self._length)[start:stop:step], self._array)

        return MappedLazySlicedPage(
            range(self._length)[start:stop:step], self._array_getter, self._mapper
        )

    def get_array(self) -> Sequence[_T]:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        array = self._array
        if array is None:
            array = self._array_getter()
            array = tuple(map(self._mapper, array))
            self._array = array
            self._patch(array)

        return array

    def copy(
        self, copier: Callable[[Any], Any], mapper: Callable[[Any], Any]
    ) -> "Union[MappedPage[_T], MappedLazyPage[_T]]":
        """Return a copy of the mapped page.

        Arguments:
            copier: A callable object to convert loaded items in the source page to the copied page.
            mapper: The mapper of the new mapped page.

        Returns:
            A copy of the mapped page.

        """
        if self._array is not None:
            return MappedPage(tuple(map(copier, self._array)))

        return MappedLazyPage(self._length, self._array_getter, mapper)


class MappedLazySlicedPage(MappedPageBase[_T]):
    """LazySlicedPage with a mapper for converting every item in the source array.

    Arguments:
        ranging: The range instance of this page.
        array_getter: A callable object to get the source array.
        mapper: A callable object to convert every item in the source array.

    """

    _array: Optional[Sequence[_T]] = None

    def __init__(
        self,
        ranging: range,
        array_getter: Callable[[], Sequence[_T]],
        mapper: Callable[[Any], Any],
    ) -> None:
        self._ranging = ranging
        self._array_getter = array_getter
        self._mapper = mapper

    def __len__(self) -> int:
        return len(self._ranging)

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "Union[MappedLazySlicedPage[_T], MappedSlicedPage[_T]]":
        """Return a lazy sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        if self._array is not None:
            return MappedSlicedPage(self._ranging[start:stop:step], self._array)

        return MappedLazySlicedPage(
            self._ranging[start:stop:step], self._array_getter, self._mapper
        )

    def get_array(self) -> Sequence[_T]:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        array = self._array
        if array is None:
            ranging = self._ranging
            stop = ranging.stop
            array = self._array_getter()[
                ranging.start : stop if stop != -1 else None : ranging.step
            ]
            array = tuple(map(self._mapper, array))

            self._array = array
            self._patch(array)

        return array

    def copy(
        self, copier: Callable[[Any], Any], mapper: Callable[[Any], Any]
    ) -> "Union[MappedPage[_T], MappedLazySlicedPage[_T]]":
        """Return a copy of the mapped page.

        Arguments:
            copier: A callable object to convert loaded items in the source page to the copied page.
            mapper: The mapper of the new mapped page.

        Returns:
            A copy of the mapped page.

        """
        if self._array is not None:
            return MappedPage(tuple(map(copier, self._array)))

        return MappedLazySlicedPage(self._ranging, self._array_getter, mapper)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Graviti image file class."""

from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional, Type, TypeVar, Union

import pyarrow as pa

from graviti.file.base import File, RemoteFile
from graviti.file.image_size import get_image_size
from graviti.portex import STANDARD_URL, ExternalElementResgister

if TYPE_CHECKING:
    from graviti.manager import ObjectPermissionManager


_I = TypeVar("_I", bound="Image")
_RI = TypeVar("_RI", bound="RemoteImage")


class Image(File):
    """This class represents local image files.

    Arguments:
        path: The local path of the image.

    """

    __slots__ = File.__slots__ + ("_height", "_width")

    _height: int
    _width: int

    @classmethod
    def _from_pyarrow(
        cls: Type[_I],
        scalar: pa.StructScalar,
        _: Optional["ObjectPermissionManager"] = None,
    ) -> _I:
        obj: _I = object.__new__(cls)
        pyobj = scalar.as_py()

        obj._key = pyobj["key"]
        obj._path = Path(pyobj["key"])
        obj._extension = pyobj["extension"]
        obj._size = pyobj["size"]
        obj._height = pyobj["height"]
        obj._width = pyobj["width"]

        return obj

    def _to_post_data(self) -> Dict[str, Union[int, str]]:
        post_data = super()._to_post_data()
        post_data["height"] = self.height
        post_data["width"] = self.width
        return post_data

    @property
    def height(self) -> int:
        """Get the height of the image.

        Returns:
            The height of the image.

        """
        if not hasattr(self, "_height"):
            self._height, self._width = get_image_size(self.path)
        return self._height

    @property
    def width(self) -> int:
        """Get the width of the image.

        Returns:
            The width of the image.

        """
        if not hasattr(self, "_width"):
            self._height, self._width = get_image_size(self.path)
        return self._width


@ExternalElementResgister(STANDARD_URL, "main", "file.Image", "label.Mask")
class RemoteImage(RemoteFile):
    """This class represents remote image files.

    Arguments:
        key: The key of the image file.
        extension: The extension of the image file.
        size: The size of the image file.
        height: The height of the image.
        width: The width of the image.
        object_permission_manager: The permission to access the file.

    """

    __slots__ = RemoteFile.__slots__ + ("_height", "_width")

    _height: int
    _width: int

    def __init__(  # pylint: disable=too-many-arguments
        self,
        key: str,
        extension: str,
        size: int,
        height: int,
        width: int,
        object_permission_manager: "ObjectPermissionManager",
    ) -> None:
        RemoteFile.__init__(self, key, extension, size, object_permission_manager)
        self._height = height
        self._width = width

    @classmethod
    def _from_pyarrow(  # type: ignore[override]
        cls: Type[_RI],
        scalar: pa.StructScalar,
        object_permission_manager: "ObjectPermissionManager",
    ) -> _RI:
        obj: _RI = object.__new__(cls)
        pyobj = scalar.as_py()

        obj._key = pyobj["key"]
        obj._extension = pyobj["extension"]
        obj._size = pyobj["size"]
        obj._height = pyobj["height"]
        obj._width = pyobj["width"]
        obj._object_permission = object_permission_manager

        return obj

    def _to_post_data(self) -> Dict[str, Union[int, str]]:
        post_data = super()._to_post_data()
        post_data["height"] = self.height
        post_data["width"] = self.width
        return post_data

    @property
    def height(self) -> int:
        """Get the height of the image.

        Returns:
            The height of the image.

        """
        return self._height

    @property
    def width(self) -> int:
        """Get the width of the image.

        Returns:
            The width of the image.

        """
        return self._width



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the tag."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do


def create_tag(
    access_key: str, url: str, workspace: str, dataset: str, *, name: str, revision: str
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{workspace}/{dataset}/tags`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        name: The tag name to be created for the specific commit.
        revision: The information to locate the specific commit, which can be the commit id,
            the branch name, or the tag name.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> create_tag(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     name="tag-2",
        ...     revision="986d8ea00da842ed85dd5d5cd14b5aef"
        ... )
        {
           "name": "tag-2",
           "commit_id": "986d8ea00da842ed85dd5d5cd14b5aef",
           "parent_commit_id": "a0d4065872f245e4ad1d0d1186e3d397",
           "title": "commit-1",
           "description": "",
           "committer": "graviti-example",
           "committed_at": "2021-03-03T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/tags"
    post_data = {"name": name, "revision": revision}
    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()


def list_tags(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/tags`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_tags(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST"
        ... )
        {
           "tags": [
               {
                   "name": "tag-2",
                   "commit_id": "986d8ea00da842ed85dd5d5cd14b5aef",
                   "parent_commit_id": "a0d4065872f245e4ad1d0d1186e3d397",
                   "title": "commit-1",
                   "description": "",
                   "committer": "graviti-example",
                   "committed_at": "2021-03-03T18:58:10Z"
               }
           ],
           "offset": 0,
           "record_size": 1,
           "total_count": 1
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/tags"

    params = {
        "offset": offset,
        "limit": limit,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def get_tag(access_key: str, url: str, workspace: str, dataset: str, *, tag: str) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/tags/{tag}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        tag: The name of the tag to be got.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_tag(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     tag="tag-2"
        ... )
        {
           "name": "tag-2",
           "commit_id": "986d8ea00da842ed85dd5d5cd14b5aef",
           "parent_commit_id": "a0d4065872f245e4ad1d0d1186e3d397",
           "title": "commit-1",
           "description": "",
           "committer": "graviti-example",
           "committed_at": "2021-03-03T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/tags/{tag}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def delete_tag(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    tag: str,
) -> None:
    """Execute the OpenAPI `DELETE /v2/datasets/{workspace}/{dataset}/tags/{tag}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        tag: The name of the tag to be deleted.

    Examples:
        >>> delete_tag(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     tag="tag-2"
        ... )

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/tags/{tag}"
    open_api_do("DELETE", access_key, url)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the search related Scalar."""

from typing import Any, Callable, ClassVar, Dict, Type, TypeVar, Union

import pyarrow as pa

import graviti.portex as pt
from graviti.dataframe.sql.container import _E, ArrayContainer, ScalarContainer

NUMERICAL_PRIORITIES: Dict[Type[pt.PortexType], int] = {
    pt.int32: 0,
    pt.int64: 1,
    pt.float32: 2,
    pt.float64: 3,
}

_LOM = TypeVar("_LOM", bound="LogicalOperatorsMixin")
_EOM = TypeVar("_EOM", bound="EqualOperatorsMixin")
_COM = TypeVar("_COM", bound="ComparisonOperatorsMixin")
_AOM = TypeVar("_AOM", bound="ArithmeticOperatorsMixin")
_NS = TypeVar("_NS", bound="NumberScalar")
_ES = TypeVar("_ES", bound="EnumScalar")
_TSB = TypeVar("_TSB", bound="TemporalScalarBase")


class LogicalOperatorsMixin(ScalarContainer):
    """A mixin for dynamically implementing logical operators."""

    _LOCICAL_OPERATORS: ClassVar[Dict[str, str]] = {
        "__and__": "and",
        "__or__": "or",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls._LOCICAL_OPERATORS.items():
            setattr(cls, meth, cls._get_logical_operator(opt))

    @classmethod
    def _get_logical_operator(cls: Type[_LOM], opt: str) -> Callable[[_LOM, Any], "BooleanScalar"]:
        def func(self: _LOM, other: Any) -> "BooleanScalar":
            other_expr = other.expr if isinstance(other, ScalarContainer) else other
            expr = {f"${opt}": [self.expr, other_expr]}

            return BooleanScalar(expr)

        return func


class EqualOperatorsMixin(ScalarContainer):
    """A mixin for dynamically implementing equal operators."""

    _LOCICAL_OPERATORS: ClassVar[Dict[str, str]] = {
        "__eq__": "eq",
        "__ne__": "ne",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls._LOCICAL_OPERATORS.items():
            setattr(cls, meth, cls._get_equal_operator(opt))

    @classmethod
    def _get_equal_operator(cls: Type[_EOM], opt: str) -> Callable[[_EOM, Any], "BooleanScalar"]:
        def func(self: _EOM, other: Any) -> "BooleanScalar":
            other_expr = other.expr if isinstance(other, ScalarContainer) else other
            expr = {f"${opt}": [self.expr, other_expr]}

            return BooleanScalar(expr)

        return func


class ComparisonOperatorsMixin(ScalarContainer):
    """A mixin for dynamically implementing comparison operators."""

    _COMPARISON_OPERATORS: ClassVar[Dict[str, str]] = {
        "__gt__": "gt",
        "__ge__": "gte",
        "__lt__": "lt",
        "__le__": "lte",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls._COMPARISON_OPERATORS.items():
            setattr(cls, meth, cls._get_comparison_operator(opt))

    @classmethod
    def _get_comparison_operator(
        cls: Type[_COM], opt: str
    ) -> Callable[[_COM, Any], "BooleanScalar"]:
        def func(self: _COM, other: Any) -> "BooleanScalar":
            if isinstance(other, ScalarContainer):
                if not isinstance(other, type(self)):
                    raise TypeError(
                        f"Invalid '{opt}' operation between {self.schema} and {other.schema}"
                    )
                other_expr = other.expr
            else:
                other_expr = other

            expr = {f"${opt}": [self.expr, other_expr]}
            return BooleanScalar(expr)

        return func


class ArithmeticOperatorsMixin(ScalarContainer):
    """A mixin for dynamically implementing arithmetic operators."""

    _ARITHMETIC_OPERATORS: ClassVar[Dict[str, str]] = {
        "__div__": "div",
        "__mod__": "mod",
        "__pow__": "pow",
        "__sub__": "sub",
        "__mul__": "mult",
        "__add__": "add",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls._ARITHMETIC_OPERATORS.items():
            setattr(cls, meth, cls._get_arithemtic_operator(opt))

    @classmethod
    def _get_arithemtic_operator(cls: Type[_AOM], opt: str) -> Callable[[_AOM, Any], Any]:
        raise NotImplementedError


class NumberScalar(
    LogicalOperatorsMixin, EqualOperatorsMixin, ComparisonOperatorsMixin, ArithmeticOperatorsMixin
):
    """One-dimensional array for numerical portex builtin type."""

    @classmethod
    def _get_arithemtic_operator(cls: Type[_NS], opt: str) -> Callable[[_NS, Any], _NS]:
        def func(self: _NS, other: Any) -> _NS:
            if isinstance(other, ScalarContainer):
                if not isinstance(other, type(self)):
                    raise TypeError(
                        f"Invalid '{opt}' operation between {self.schema} and {other.schema}"
                    )
                other_expr = other.expr
                schema = (
                    self.schema
                    if NUMERICAL_PRIORITIES[self.schema.__class__]
                    > NUMERICAL_PRIORITIES[other.schema.__class__]
                    else other.schema
                )
            else:
                other_expr = other
                schema = self.schema

            expr = {f"${opt}": [self.expr, other_expr]}
            return cls(expr, schema)

        return func


class BooleanScalar(LogicalOperatorsMixin, EqualOperatorsMixin):
    """One-dimensional array for portex builtin type boolean."""

    def __init__(self, expr: _E) -> None:
        super().__init__(expr, pt.boolean())


class StringScalar(LogicalOperatorsMixin, EqualOperatorsMixin):
    """One-dimensional array for portex builtin type string."""


class EnumScalar(EqualOperatorsMixin):
    """One-dimensional array for portex builtin type enum."""

    @classmethod
    def _get_equal_operator(cls: Type[_ES], opt: str) -> Callable[[_ES, Any], "BooleanScalar"]:
        def func(self: _ES, other: Any) -> "BooleanScalar":
            if isinstance(other, ScalarContainer):
                other_expr: Any = other.expr
            else:
                enum: pt.enum = self.schema.to_builtin()  # type: ignore[assignment]
                other_expr = enum.values.value_to_index[other]

            expr = {f"${opt}": [self.expr, other_expr]}
            return BooleanScalar(expr)

        return func


class TemporalScalarBase(EqualOperatorsMixin, ComparisonOperatorsMixin):
    """One-dimensional array for portex builtin temporal type."""

    def _time_to_int(self, value: Any) -> int:
        return pa.scalar(value, self.schema.to_pyarrow()).value  # type: ignore[no-any-return]

    @classmethod
    def _get_equal_operator(cls: Type[_TSB], opt: str) -> Callable[[_TSB, Any], "BooleanScalar"]:
        def func(self: _TSB, other: Any) -> "BooleanScalar":
            if isinstance(other, ScalarContainer):
                other_expr: Any = other.expr
            else:
                other_expr = self._time_to_int(other)  # pylint: disable=protected-access

            expr = {f"${opt}": [self.expr, other_expr]}
            return BooleanScalar(expr)

        return func

    @classmethod
    def _get_comparison_operator(
        cls: Type[_TSB], opt: str
    ) -> Callable[[_TSB, Any], "BooleanScalar"]:
        def func(self: _TSB, other: Any) -> "BooleanScalar":
            if isinstance(other, ScalarContainer):
                if not isinstance(other, type(self)):
                    raise TypeError(
                        f"Invalid '{opt}' operation between {self.schema} and {other.schema}"
                    )
                other_expr: Any = other.expr
            else:
                other_expr = self._time_to_int(other)  # pylint: disable=protected-access

            expr = {f"${opt}": [self.expr, other_expr]}
            return BooleanScalar(expr)

        return func


class DateScalar(TemporalScalarBase):
    """One-dimensional array for portex builtin date type."""

    def _time_to_int(self, value: Any) -> int:
        scalar = pa.scalar(value, self.schema.to_pyarrow())
        return scalar.cast(pa.int32()).as_py()  # type: ignore[no-any-return]


class TimeScalar(TemporalScalarBase):
    """One-dimensional array for portex builtin time type."""


class TimestampScalar(TemporalScalarBase):
    """One-dimensional array for portex builtin timestamp type."""


class TimedeltaScalar(TemporalScalarBase):
    """One-dimensional array for portex builtin timedelta type."""


class RowSeries(ScalarContainer):
    """The One-dimensional array for the search."""

    schema: pt.PortexRecordBase

    def __init__(self, schema: pt.PortexRecordBase) -> None:
        super().__init__("$", schema)

    def __getitem__(self, key: str) -> Union[ArrayContainer, ScalarContainer]:
        field = self.schema[key]
        return field.search_container.item_container.from_upper(f"{self.expr}.{key}", field)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti indexing related class."""

from typing import TYPE_CHECKING, Any, Iterable, Tuple, Union, overload

from graviti.dataframe.column.series import Series
from graviti.openapi import RECORD_KEY
from graviti.operation import UpdateData
from graviti.utility import NestedDict

if TYPE_CHECKING:
    from graviti.dataframe import DataFrame
    from graviti.dataframe.row.series import Series as RowSeries


class DataFrameILocIndexer:
    """Index class for DataFrame.iloc."""

    def __init__(self, obj: "DataFrame") -> None:
        self.obj = obj

    # @overload
    # def __getitem__(self, key: Iterable[int]) -> "DataFrame":
    #     ...

    @overload
    def __getitem__(self, key: slice) -> "DataFrame":
        ...

    @overload
    def __getitem__(self, key: Tuple[Union[int, slice], str]) -> Any:
        ...

    @overload
    def __getitem__(self, key: int) -> "RowSeries":
        ...

    def __getitem__(
        self, key: Union[int, slice, Tuple[Union[int, slice], str]]
    ) -> Union[Any, "DataFrame", "RowSeries"]:
        if isinstance(key, tuple):
            index, name = key
            return self.obj[name].iloc[index]

        if isinstance(key, int):
            return self.obj._get_item_by_location(key)

        return self.obj._get_slice_by_location(key, self.obj.schema.copy())

    @overload
    def __setitem__(self, key: int, value: NestedDict[str, Any]) -> None:
        ...

    @overload
    def __setitem__(
        self, key: slice, value: Union[Iterable[NestedDict[str, Any]], "DataFrame"]
    ) -> None:
        ...

    @overload
    def __setitem__(self, key: Tuple[int, str], value: Any) -> None:
        ...

    @overload
    def __setitem__(self, key: Tuple[slice, str], value: Union[Iterable[Any], Series]) -> None:
        ...

    def __setitem__(
        self,
        key: Union[int, slice, Tuple[int, str], Tuple[slice, str]],
        value: Union[
            NestedDict[str, Any],
            "DataFrame",
            Iterable[NestedDict[str, Any]],
            Any,
            Iterable[Any],
            Series,
        ],
    ) -> None:
        if isinstance(key, tuple):
            index, name = key
            self.obj[name].loc[index] = value
            return

        if isinstance(key, int):
            # pylint: disable=protected-access
            value = self.obj._from_pyarrow(
                self.obj._pylist_to_pyarrow([value], self.obj.schema),
                self.obj.schema,
            )
            key = slice(key, key + 1)
        elif not isinstance(value, self.obj.__class__):
            value = self.obj._from_pyarrow(
                self.obj._pylist_to_pyarrow(value, self.obj.schema),  # type: ignore[arg-type]
                self.obj.schema,
            )
        elif not self.obj.schema.to_pyarrow().equals(value.schema.to_pyarrow()):
            raise TypeError("The schema of the given DataFrame is mismatched")

        self.obj._set_slice_by_location(key, value)
        if self.obj.operations is not None:
            df = value.copy()
            df[RECORD_KEY] = self.obj._record_key[key]  # type: ignore[index, assignment]
            self.obj.operations.append(UpdateData(df))

    def __delitem__(self, key: Union[int, slice]) -> None:
        if self.obj._root is not None:
            raise TypeError(
                "'iloc.__delitem__' is not supported for the DataFrame"
                "which is a member of another DataFrame"
            )

        self.obj._del_item_by_location(key)


class DataFrameLocIndexer:
    """Index class for DataFrame.loc."""

    def __init__(self, obj: "DataFrame") -> None:
        self.obj: "DataFrame" = obj

    # @overload
    # def __getitem__(self, key: Iterable[int]) -> "DataFrame":
    #     ...

    @overload
    def __getitem__(self, key: slice) -> "DataFrame":
        ...

    @overload
    def __getitem__(self, key: Tuple[Union[int, slice], str]) -> Any:
        ...

    @overload
    def __getitem__(self, key: int) -> "RowSeries":
        ...

    def __getitem__(
        self, key: Union[int, slice, Tuple[Union[int, slice], str]]
    ) -> Union[Any, "DataFrame", "RowSeries"]:
        if isinstance(key, tuple):
            index, name = key
            return self.obj[name].loc[index]

        if isinstance(key, int):
            return self.obj._get_item_by_location(key)

        return self.obj._get_slice_by_location(key, self.obj.schema.copy())

    @overload
    def __setitem__(self, key: int, value: NestedDict[str, Any]) -> None:
        ...

    @overload
    def __setitem__(
        self, key: slice, value: Union[Iterable[NestedDict[str, Any]], "DataFrame"]
    ) -> None:
        ...

    @overload
    def __setitem__(self, key: Tuple[int, str], value: Any) -> None:
        ...

    @overload
    def __setitem__(self, key: Tuple[slice, str], value: Union[Iterable[Any], Series]) -> None:
        ...

    def __setitem__(
        self,
        key: Union[int, slice, Tuple[int, str], Tuple[slice, str]],
        value: Union[
            NestedDict[str, Any],
            "DataFrame",
            Iterable[NestedDict[str, Any]],
            Any,
            Iterable[Any],
            Series,
        ],
    ) -> None:
        if isinstance(key, tuple):
            index, name = key
            self.obj[name].iloc[index] = value
            return

        if isinstance(key, int):
            # pylint: disable=protected-access
            value = self.obj._from_pyarrow(
                self.obj._pylist_to_pyarrow([value], self.obj.schema),
                self.obj.schema,
            )
            key = slice(key, key + 1)
        elif not isinstance(value, self.obj.__class__):
            value = self.obj._from_pyarrow(
                self.obj._pylist_to_pyarrow(value, self.obj.schema),  # type: ignore[arg-type]
                self.obj.schema,
            )
        elif not self.obj.schema.to_pyarrow().equals(value.schema.to_pyarrow()):
            raise TypeError("The schema of the given DataFrame is mismatched")

        self.obj._set_slice_by_location(key, value)
        if self.obj.operations is not None:
            df = value.copy()
            df[RECORD_KEY] = self.obj._record_key[key]  # type: ignore[index, assignment]
            self.obj.operations.append(UpdateData(df))

    def __delitem__(self, key: Union[int, slice]) -> None:
        if self.obj._root is not None:
            raise TypeError(
                "'loc.__delitem__' is not supported for the DataFrame"
                "which is a member of another DataFrame"
            )

        self.obj._del_item_by_location(key)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Branch and BranchManager."""

from typing import TYPE_CHECKING, Generator

from graviti.exception import NoCommitsError, ResourceNameError
from graviti.manager.commit import NamedCommit
from graviti.manager.common import CURRENT_COMMIT, LIMIT, check_head_status
from graviti.manager.lazy import LazyPagingList
from graviti.openapi import create_branch, delete_branch, get_branch, list_branches
from graviti.utility import check_type

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset


class Branch(NamedCommit):
    """This class defines the structure of a branch.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        name: The name of the branch.
        commit_id: The commit id.

    """


class BranchManager:
    """This class defines the operations on the branch in Graviti.

    Arguments:
        dataset: :class:`~graviti.manager.dataset.Dataset` instance.

    """

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _generate(self, offset: int, limit: int) -> Generator[Branch, None, int]:
        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = list_branches(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            offset=offset,
            limit=limit,
        )

        head = _dataset.HEAD
        for item in response["branches"]:
            check_head_status(head, item["name"], item["commit_id"])
            yield Branch.from_response(_dataset, item)

        return response["total_count"]  # type: ignore[no-any-return]

    def create(self, name: str, revision: str = CURRENT_COMMIT) -> Branch:
        """Create a branch.

        Arguments:
            name: The branch name.
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name. The default value is the current commit of the
                dataset.

        Raises:
            NoCommitsError: When create branches on default branch without commit history.

        Returns:
            The :class:`.Branch` instance with the given name.

        """
        _dataset = self._dataset
        _workspace = _dataset.workspace
        head = _dataset.HEAD
        if revision is CURRENT_COMMIT:
            _revision = head.commit_id
            if _revision is None:
                raise NoCommitsError(
                    "Creating branches on the default branch without commit history is not allowed."
                    "Please commit a draft first"
                )
        else:
            _revision = revision

        response = create_branch(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            name=name,
            revision=_revision,
        )

        check_head_status(head, _revision, response["commit_id"])
        return Branch.from_response(_dataset, response)

    def get(self, name: str) -> Branch:
        """Get the branch with the given name.

        Arguments:
            name: The required branch name.

        Raises:
            ResourceNameError: When the given name is an empty string.

        Returns:
            The :class:`.Branch` instance with the given name.

        """
        check_type("name", name, str)
        if not name:
            raise ResourceNameError("branch", name)

        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = get_branch(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            branch=name,
        )
        check_head_status(_dataset.HEAD, name, response["commit_id"])

        return Branch.from_response(_dataset, response)

    def list(self) -> LazyPagingList[Branch]:
        """List the information of branches.

        Returns:
            The LazyPagingList of :class:`branches<.Branch>` instances.

        """
        return LazyPagingList(self._generate, LIMIT)

    def delete(self, name: str) -> None:
        """Delete a branch.

        Arguments:
            name: The name of the branch to be deleted.

        Raises:
            ResourceNameError: When the given name is an empty string.

        """
        check_type("name", name, str)
        if not name:
            raise ResourceNameError("branch", name)

        _workspace = self._dataset.workspace
        delete_branch(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self._dataset.name,
            branch=name,
        )



# !/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of logging utilities."""

import json
from typing import MutableMapping

from requests.models import PreparedRequest, Response
from requests_toolbelt.multipart.encoder import FileWrapper, MultipartEncoder

REQUEST_TEMPLATE = """
===================================================================
########################## HTTP Request ###########################
"url": {}
"method": {}
"headers": {}
"body": {}
"""

RESPONSE_TEMPLATE = """
########################## HTTP Response ##########################
"url": {}
"status_code": {}
"reason": {}
"headers": {}
"content": {}
"cost_time": {}s
===================================================================
"""


class RequestLogging:
    """This class used to lazy load request to logging.

    Arguments:
        request: The request of the request.

    """

    def __init__(self, request: PreparedRequest) -> None:
        self._request = request

    def __str__(self) -> str:
        return _dump_request(self._request)


class ResponseLogging:
    """This class used to lazy load response to logging.

    Arguments:
        response: The response of the request.

    """

    def __init__(self, response: Response) -> None:
        self._response = response

    def __str__(self) -> str:
        return dump_request_and_response(self._response)


def dump_request_and_response(response: Response) -> str:
    r"""Dumps http request and response.

    Arguments:
        response: Http response and response.

    Returns:
        Http request and response for logging, sample::

            ===================================================================
            ########################## HTTP Request ###########################
            "url": https://gas.graviti.cn/gatewayv2/content-store/putObject
            "method": POST
            "headers": {
              "User-Agent": "python-requests/2.23.0",
              "Accept-Encoding": "gzip, deflate",
              "Accept": "*/*",
              "Connection": "keep-alive",
              "X-Token": "c3b1808b21024eb38f066809431e5bb9",
              "Content-Type": "multipart/form-data; boundary=5adff1fc0524465593d6a9ad68aad7f9",
              "Content-Length": "330001"
            }
            "body":
            --5adff1fc0524465593d6a9ad68aad7f9
            b'Content-Disposition: form-data; name="contentSetId"\r\n\r\n'
            b'e6110ff1-9e7c-4c98-aaf9-5e35522969b9'

            --5adff1fc0524465593d6a9ad68aad7f9
            b'Content-Disposition: form-data; name="filePath"\r\n\r\n'
            b'4.jpg'

            --5adff1fc0524465593d6a9ad68aad7f9
            b'Content-Disposition: form-data; name="fileData"; filename="4.jpg"\r\n\r\n'
            [329633 bytes of object data]

            --5adff1fc0524465593d6a9ad68aad7f9--

            ########################## HTTP Response ###########
            "url": https://gas.graviti.cn/gatewayv2/content-stor
            "status_code": 200
            "reason": OK
            "headers": {
              "Date": "Sat, 23 May 2020 13:05:09 GMT",
              "Content-Type": "application/json;charset=utf-8",
              "Content-Length": "69",
              "Connection": "keep-alive",
              "Access-Control-Allow-Origin": "*",
              "X-Kong-Upstream-Latency": "180",
              "X-Kong-Proxy-Latency": "112",
              "Via": "kong/2.0.4"
            }
            "content": {
              "success": true,
              "code": "DATACENTER-0",
              "message": "success",
              "data": {}
            }
            "cost_time": 0.0813691616058
            ====================================================

    """
    return _dump_request(response.request) + _dump_response(response)


def _dump_request(request: PreparedRequest) -> str:
    """Dump http request.

    Arguments:
        request: Http request.

    Returns:
        String of http request for logging.

    """
    headers = _dump_headers(request.headers)
    body = "N/A"
    if "Content-Type" in request.headers:
        if request.headers["Content-Type"].startswith("multipart/form-data"):
            body = _dump_multipart_encoder(request.body)
        elif isinstance(request.body, bytes):
            body = request.body.decode("unicode_escape")

    return REQUEST_TEMPLATE.format(request.url, request.method, headers, body)


def _dump_response(response: Response) -> str:
    """Dump http response.

    Arguments:
        response: Http response.

    Returns:
        String of http response for logging.

    """
    headers = _dump_headers(response.headers)
    content = "N/A"

    if "Content-Type" in response.headers:
        content_type = response.headers["Content-Type"]
        if content_type.startswith("application/json"):
            content = json.dumps(response.json(), indent=2)
        elif content_type.startswith("text"):
            content = response.text
        elif len(response.content) > 512:
            content = f"[{len(response.content)} bytes of object data]"
        else:
            content = str(response.content)

    return RESPONSE_TEMPLATE.format(
        response.url,
        response.status_code,
        response.reason,
        headers,
        content,
        response.elapsed.total_seconds(),
    )


def _dump_multipart_encoder(body: MultipartEncoder) -> str:
    """Dump MultipartEncoder multipart/form-data Content-Type post_data.

    Arguments:
        body: MultipartEncoder multipart/formdata post data.

    Returns:
        String of multipart/form-data post data for logging.

    """
    lines = []
    boundary = body.boundary
    for part in body.parts:
        lines.append("\n" + boundary)
        lines.append(str(part.headers))
        body = part.body
        if isinstance(body, FileWrapper):
            if body.fd.closed:
                lines.append("[file closed]")
            else:
                lines.append(f"[{body.fd.tell()} bytes of object data]")
        else:
            lines.append(str(body.getvalue()))

    lines.append("\n" + boundary + "--")

    return "\n".join(lines)


def _dump_headers(headers: MutableMapping[str, str]) -> str:
    """Dump http headers as json format string for logging.

    Arguments:
        headers: Headers of http request or response.

    Returns:
        Json format string of headers content.

    """
    return json.dumps(dict(headers), indent=2)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The search container and register."""

from typing import TYPE_CHECKING, Any, ClassVar, Dict, Type, TypeVar, Union

import graviti.portex as pt

if TYPE_CHECKING:
    from graviti.portex.base import PortexType


_S = TypeVar("_S", bound="ScalarContainer")
_A = TypeVar("_A", bound="ArrayContainer")
_E = Union[str, Dict[str, Any]]


class ScalarContainer:
    """The base class for the search scalar container.

    Arguments:
        expr: The expression of the search.
        schema: The schema of the series.

    """

    expr: _E
    schema: pt.PortexType

    def __init__(self, expr: _E, schema: pt.PortexType) -> None:
        self.expr = expr
        self.schema = schema

    @classmethod
    def from_upper(cls: Type[_S], expr: _E, schema: pt.PortexType) -> _S:
        """Instantiate a Search object from the upper level.

        Arguments:
            expr: The expression of the search.
            schema: The schema of the series.

        Returns:
            The loaded object.

        """
        obj: _S = object.__new__(cls)
        obj.expr = expr
        obj.schema = schema

        return obj


class ArrayContainer:
    """The base class for the search array container.

    Arguments:
        expr: The expression of the search.
        schema: The schema of the series.
        upper_expr: The expression of the search.

    """

    prefix: str
    item_container: ClassVar[Type[Union[ScalarContainer, "ArrayContainer"]]]

    def __init__(self, expr: _E, schema: pt.PortexType, upper_expr: _E) -> None:
        self.expr = expr
        self.schema = schema
        self.upper_expr = upper_expr

    @classmethod
    def from_upper(cls, expr: _E, schema: pt.PortexType) -> "ArrayContainer":
        """Instantiate a Search object from the upper level.

        Arguments:
            expr: The upper expression of the search.
            schema: The schema of the series.

        Returns:
            The loaded object.

        """
        return cls(cls.prefix, schema, expr)


class SearchContainerRegister:
    """The class decorator to connect portex type and the search array container.

    Arguments:
        portex_types: The portex types needs to be connected.

    """

    def __init__(self, *portex_types: Type["PortexType"]) -> None:
        self._portex_types = portex_types

    def __call__(self, container: Type[_A]) -> Type[_A]:
        """Connect search array container with the portex types.

        Arguments:
            container: The search array container needs to be connected.

        Returns:
            The input container class unchanged.

        """
        for portex_type in self._portex_types:
            portex_type.search_container = container

        return container



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Portex type builder related classes."""


import os
from hashlib import md5
from pathlib import Path
from shutil import rmtree
from subprocess import PIPE, CalledProcessError, run
from tempfile import gettempdir
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Type, TypeVar

import yaml

from graviti.exception import GitCommandError, GitNotFoundError
from graviti.portex import ptype as PTYPE
from graviti.portex.base import PortexRecordBase
from graviti.portex.external import PortexExternalType
from graviti.portex.factory import ConnectedFieldsFactory, TypeFactory
from graviti.portex.package import ExternalPackage, Imports, packages
from graviti.portex.param import Param, Params
from graviti.portex.register import ExternalContainerRegister

if TYPE_CHECKING:
    from subprocess import CompletedProcess

    from graviti.portex.base import PortexType


EXTERNAL_TYPE_TO_CONTAINER = ExternalContainerRegister.EXTERNAL_TYPE_TO_CONTAINER

_I = TypeVar("_I", bound="BuilderImports")


class PackageRepo:
    """The local git repo of the external Portex package.

    Arguments:
        url: The git repo url of the external package.
        revision: The git repo revision (tag/commit) of the external package.

    """

    _env: Dict[str, Any] = {
        k: v
        for k, v in os.environ.items()
        if not k.startswith("GIT_")
        or k.startswith(("GIT_CONFIG_KEY_", "GIT_CONFIG_VALUE_"))
        or k
        in {
            "GIT_EXEC_PATH",
            "GIT_SSH",
            "GIT_SSH_COMMAND",
            "GIT_SSL_CAINFO",
            "GIT_SSL_NO_VERIFY",
            "GIT_CONFIG_COUNT",
            "GIT_HTTP_PROXY_AUTHMETHOD",
        }
    }

    def __init__(self, url: str, revision: str) -> None:
        tempdir = Path(gettempdir()) / "portex"
        tempdir.mkdir(exist_ok=True)

        md5_instance = md5()
        md5_instance.update(url.encode("utf-8"))
        md5_instance.update(revision.encode("utf-8"))

        self._path = tempdir / md5_instance.hexdigest()
        self._url = url
        self._revision = revision

        try:
            self._prepare_repo()
        except FileNotFoundError:
            raise GitNotFoundError() from None

    def _prepare_repo(self) -> None:
        if not self._path.exists():
            self._clone_repo()
        elif not self._check_repo_integrity():
            rmtree(self._path)
            self._clone_repo()

    def _run(self, args: List[str]) -> "CompletedProcess[bytes]":
        return run(args, cwd=self._path, env=self._env, stdout=PIPE, stderr=PIPE, check=True)

    def _init_repo(self) -> None:
        self._run(["git", "init"])
        self._run(["git", "remote", "add", "origin", self._url])

    def _shallow_fetch(self) -> None:
        self._run(["git", "fetch", "origin", self._revision, "--depth=1"])
        self._run(["git", "checkout", "FETCH_HEAD"])

    def _deep_fetch(self) -> None:
        try:
            self._run(["git", "fetch", "origin"])
        except CalledProcessError as error:
            raise GitCommandError(
                "'git fetch' failed, most likely due to the repo url is invalid.",
                error,
            ) from None

        try:
            self._run(["git", "checkout", self._revision])
        except CalledProcessError as error:
            raise GitCommandError(
                "'git checkout' failed, most likely due to the repo revision is invalid.",
                error,
            ) from None

    def _check_repo_integrity(self) -> bool:
        try:
            result = self._run(["git", "status", "--porcelain"])
        except CalledProcessError:
            # The git command failed means the git repo has been cleaned or broken
            return False

        return not bool(result.stdout)

    def _clone_repo(self) -> None:
        print(f"Cloning repo '{self._url}@{self._revision}'")

        path = self._path
        path.mkdir()
        try:
            self._init_repo()
            try:
                self._shallow_fetch()
            except CalledProcessError:
                self._deep_fetch()
        except (CalledProcessError, GitCommandError, FileNotFoundError, KeyboardInterrupt):
            rmtree(path)
            raise

        print(f"Cloned to '{path}'")

    def get_root(self) -> Path:
        """Get the root directory path of the package repo.

        Returns:
            The root directory path of the package repo.

        Raises:
            TypeError: when the "ROOT.yaml" not found or more than one "ROOT.yaml" found.

        """
        roots = list(self._path.glob("**/ROOT.yaml"))

        if len(roots) == 0:
            raise TypeError("No 'ROOT.yaml' file found")
        if len(roots) >= 2:
            raise TypeError("More than one 'ROOT.yaml' file found")

        return roots[0].parent


class PackageBuilder:
    """The builder of the external Portex package.

    Arguments:
        url: The git repo url of the external package.
        revision: The git repo revision (tag/commit) of the external package.

    """

    def __init__(self, url: str, revision: str) -> None:
        self.package = ExternalPackage(url, revision)

        self._builders = self._create_type_builders()

    def __getitem__(self, key: str) -> Type["PortexExternalType"]:
        try:
            return self.package[key]
        except KeyError:
            return self._builders.__getitem__(key).build()

    def _create_type_builders(self) -> Dict[str, "TypeBuilder"]:
        repo = PackageRepo(self.package.url, self.package.revision)
        root = repo.get_root()

        builders = {}
        for yaml_file in root.glob("**/*.yaml"):
            if yaml_file.name == "ROOT.yaml":
                continue

            parts = (*yaml_file.relative_to(root).parent.parts, yaml_file.stem)
            name = ".".join(parts)
            builders[name] = TypeBuilder(name, yaml_file, self)

        return builders

    def build(self) -> ExternalPackage:
        """Build the Portex external package.

        Returns:
            The builded Portex external package.

        """
        for builder in self._builders.values():
            if builder.is_building:
                continue
            builder.build()

        return self.package


class TypeBuilder:
    """The builder of the external Portex template type.

    Arguments:
        name: The name of the Portex template type.
        path: The source file path of the Portex template type.
        package: The package the Portex template type belongs to.

    """

    def __init__(self, name: str, path: Path, builder: PackageBuilder) -> None:
        self._name = name
        self._path = path
        self._builder = builder
        self.is_building = False

    def build(self) -> Type["PortexExternalType"]:
        """Build the Portex external type.

        Returns:
            The builded Portex external type.

        Raises:
            TypeError: Raise when circular reference detected.

        """
        if self.is_building:
            raise TypeError("Circular reference")

        self.is_building = True

        with self._path.open() as fp:
            content = yaml.load(fp, yaml.Loader)

        params_pyobj = content.get("parameters", [])
        decl = content["declaration"]

        imports = BuilderImports.from_pyobj(content.get("imports", []), self._builder)

        factory = TypeFactory(decl, imports)

        keys = factory.keys
        params = Params.from_pyobj(params_pyobj, keys)

        params.add(Param("nullable", False, ptype=PTYPE.Boolean))

        class_attrs: Dict[str, Any] = {
            "__module__": __name__,
            "_factory": factory,
            "params": params,
            "package": self._builder.package,
        }
        if issubclass(factory.class_, PortexRecordBase):
            bases: Tuple[Type["PortexType"], ...] = (PortexRecordBase, PortexExternalType)
            class_attrs["_fields_factory"] = ConnectedFieldsFactory(
                decl, factory.class_, imports, factory.transform_kwargs
            )
        else:
            bases = (PortexExternalType,)
        type_ = type(self._name, bases, class_attrs)

        self._builder.package[self._name] = type_

        return type_


class BuilderImports(Imports):
    """The imports of the Portex template type.

    Arguments:
        package: The package the portex belongs to.

    """

    _builder: PackageBuilder

    def __getitem__(self, key: str) -> Type["PortexType"]:
        try:
            return super().__getitem__(key)
        except KeyError:
            return self._builder.__getitem__(key)

    @classmethod
    def from_pyobj(  # type: ignore[override] # pylint: disable=arguments-differ
        cls: Type[_I], content: List[Dict[str, Any]], builder: PackageBuilder
    ) -> _I:
        """Create :class:`Imports` instance from python list.

        Arguments:
            content: A python list representing imported types.
            builder: The package builder.

        Returns:
            A :class:`Imports` instance created from the input python list.

        """
        imports = super().from_pyobj(content)
        imports._builder = builder  # pylint: disable=protected-access
        return imports


def build_package(url: str, revision: str) -> ExternalPackage:
    """Build an external package.

    Arguments:
        url: The git repo url of the external package.
        revision: The git repo revision (tag/commit) of the external package.

    Returns:
        The :class:`ExternalPackage` instance.

    """
    builder = PackageBuilder(url, revision)
    package = builder.build()
    packages.externals[url, revision] = package
    return package



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""The portex type register related classes."""


from typing import TYPE_CHECKING, Any, ClassVar, Dict, Tuple, Type, TypeVar

from graviti.portex.package import packages
from graviti.utility import urlnorm

if TYPE_CHECKING:
    from graviti.dataframe import Container
    from graviti.portex.base import PortexType

_T = TypeVar("_T")
_P = TypeVar("_P", bound="PortexType")
_C = TypeVar("_C", bound="Container")

STANDARD_URL = "https://github.com/Project-OpenBytes/portex-standard"


class ContainerRegister:
    """The class decorator to connect portex type and the data container.

    Arguments:
        portex_types: The portex types needs to be connected.

    """

    def __init__(self, *portex_types: Type["PortexType"]) -> None:
        self._portex_types = portex_types

    def __call__(self, container: Type[_C]) -> Type[_C]:
        """Connect data container with the portex types.

        Arguments:
            container: The data container needs to be connected.

        Returns:
            The input container class unchanged.

        """
        for portex_type in self._portex_types:
            portex_type.container = container

        return container


class ExternalContainerRegister:
    """The class decorator to connect portex external type and the data container.

    Arguments:
        url: The git repo url of the external package.
        revision: The git repo revision (tag/commit) of the external package.
        name: The portex external type name.

    """

    EXTERNAL_TYPE_TO_CONTAINER: ClassVar[Dict[Tuple[str, str, str], Type["Container"]]] = {}

    def __init__(self, url: str, revision: str, *names: str) -> None:
        self._url = urlnorm(url)
        self._revision = revision
        self._names = names

    def __call__(self, container: Type[_C]) -> Type[_C]:
        """Connect data container with the portex external types.

        Arguments:
            container: The data container needs to be connected.

        Returns:
            The input container class unchanged.

        """
        package = packages.externals.get((self._url, self._revision))

        for name in self._names:
            self.EXTERNAL_TYPE_TO_CONTAINER[self._url, self._revision, name] = container

            if package:
                package[name].container = container

        return container


class ExternalElementResgister:
    """The class decorator to connect portex external type and the element class.

    Arguments:
        url: The git repo url of the external package.
        revision: The git repo revision (tag/commit) of the external package.
        name: The portex external type name.

    """

    EXTERNAL_TYPE_TO_ELEMENT: ClassVar[Dict[Tuple[str, str, str], Type[Any]]] = {}

    def __init__(self, url: str, revision: str, *names: str) -> None:
        self._url = urlnorm(url)
        self._revision = revision
        self._names = names

    def __call__(self, element: Type[_T]) -> Type[_T]:
        """Connect element class with the portex external types.

        Arguments:
            element: The elemenet class needs to be connected.

        Returns:
            The input element class unchanged.

        """
        package = packages.externals.get((self._url, self._revision))

        for name in self._names:
            self.EXTERNAL_TYPE_TO_ELEMENT[self._url, self._revision, name] = element

            if package:
                package[name].element = element

        return element


class PyArrowConversionRegister:
    """Register the Portex type to set the conversion from PyArrow to Portex.

    Arguments:
        pyarrow_type_ids: The id of the corresponding PyArrow types.

    """

    PYARROW_TYPE_ID_TO_PORTEX_TYPE: ClassVar[Dict[int, Type["PortexType"]]] = {}

    def __init__(self, *pyarrow_type_ids: int) -> None:
        self._pyarrow_type_ids = pyarrow_type_ids

    def __call__(self, portex_type: Type[_P]) -> Type[_P]:
        """Register the Portex type and return it back.

        Arguments:
            portex_type: The Portex type to register.

        Returns:
            The original Portex type.

        """
        mapping = self.PYARROW_TYPE_ID_TO_PORTEX_TYPE
        for pyarrow_type_id in self._pyarrow_type_ids:
            mapping[pyarrow_type_id] = portex_type

        return portex_type



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Basic concepts of Graviti custom exceptions."""

from subprocess import CalledProcessError
from typing import Dict, Optional, Tuple, Type, TypeVar

from requests.models import Response


class GravitiException(Exception):
    """This is the base class for Graviti custom exceptions.

    Arguments:
       message: The error message.

    """

    def __init__(self, message: Optional[str] = None):
        super().__init__()
        self._message = message

    def __str__(self) -> str:
        return self._message if self._message else ""


class UtilityError(GravitiException):
    """This is the base class for custom exceptions in Graviti utility module."""


class ImageDecodeError(UtilityError):
    """This class defines the exception for the image decode errors."""


class PortexError(GravitiException):
    """This is the base class for custom exceptions in Graviti portex module."""


class FieldNameConflictError(PortexError):
    """This class defines the exception for the portex field name error."""


class GitNotFoundError(PortexError):
    """This class defines the exception for the git command not found error.

    Arguments:
       message: The error message.

    """

    _MESSAGE = (
        "'git' command failed, most likely due to the 'git' is not installed.\n"
        "Please install 'git' first: https://git-scm.com/book/en/v2/Getting-Started-Installing-Git"
    )

    def __init__(self, message: str = _MESSAGE):
        super().__init__(message)


class GitCommandError(PortexError):
    """This class defines the exception for the git command related error.

    Arguments:
       message: The error message.
       called_process_error: The CalledProcessError raised from the subprocess.run().

    """

    def __init__(self, message: str, called_process_error: CalledProcessError):
        super().__init__(message)
        self.called_process_error = called_process_error

    def __str__(self) -> str:
        error = self.called_process_error
        return (
            f"{self._message}\n"
            f"  command: {error.cmd}\n"
            f"  returncode: {error.returncode}\n"
            f"  stdout: {error.stdout.decode()}\n"
            f"  stderr: {error.stderr.decode()}"
        )


class OperationError(GravitiException):
    """This is the base class for custom exceptions in Graviti operation module."""


class ObjectCopyError(OperationError):
    """This class defines the exception for object copy error."""


class ManagerError(GravitiException):
    """This is the base class for custom exceptions in Graviti manager module."""


class StatusError(ManagerError):
    """This class defines the exception for illegal status."""


class NoCommitsError(StatusError):
    """This class defines the exception for illegal operations on dataset with no commit history."""


class ResourceNameError(ManagerError):
    """This class defines the exception for invalid resource names."""

    def __init__(self, resource: str, name: str) -> None:
        super().__init__()
        self._resource = resource
        self._name = name

    def __str__(self) -> str:
        return f'The {self._resource} name "{self._name}" is invalid'


class ResponseError(ManagerError):
    """This class defines the exception for post response error.

    Arguments:
        response: The response of the request.

    Attributes:
        response: The response of the request.

    """

    # https://github.com/python/mypy/issues/6473
    _INDENT = " " * len(__qualname__)  # type: ignore[name-defined]

    STATUS_CODE: int
    ERROR_CODE: Optional[str]

    def __init__(
        self, message: Optional[str] = None, *, response: Optional[Response] = None
    ) -> None:
        super().__init__(message)
        if response is not None:
            self.response = response

    def __init_subclass__(cls) -> None:
        cls._INDENT = " " * len(cls.__name__)

    def __str__(self) -> str:
        if hasattr(self, "response"):
            return (
                f"Unexpected status code({self.response.status_code})! {self.response.url}!"
                f"\n{self._INDENT}  {self.response.text}"
            )
        return super().__str__()


_R = TypeVar("_R", bound=Type[ResponseError])


class ResponseErrorRegister:
    """A class decorator to register the ResponseError into the distributor.

    Arguments:
        status_code: The http status code of the specific ResponseError.
        error_code: The response error code of the specific ResponseError.

    """

    RESPONSE_ERROR_DISTRIBUTOR: Dict[Tuple[int, Optional[str]], Type[ResponseError]] = {}

    def __init__(self, status_code: int, error_code: Optional[str] = None) -> None:
        self._status_code = status_code
        self._error_code = error_code

    def __call__(self, response_error: _R) -> _R:
        """Register the ResponseError into the distributor.

        Arguments:
            response_error: The response error needs to be registered.

        Returns:
            The input response error class unchanged.

        """
        response_error.STATUS_CODE = self._status_code
        response_error.ERROR_CODE = self._error_code

        self.RESPONSE_ERROR_DISTRIBUTOR[(self._status_code, self._error_code)] = response_error

        return response_error


@ResponseErrorRegister(403, "AccessDenied")
class AccessDeniedError(ResponseError):
    """This class defines the exception for access denied response error."""


@ResponseErrorRegister(403, "Forbidden")
class ForbiddenError(ResponseError):
    """This class defines the exception for illegal operations Graviti forbids."""


@ResponseErrorRegister(400, "InvalidParamsValue")
class InvalidParamsError(ResponseError):
    """This class defines the exception for invalid parameters response error."""


@ResponseErrorRegister(409, "NameConflict")
class NameConflictError(ResponseError):
    """This class defines the exception for name conflict response error."""


@ResponseErrorRegister(400, "RequestParamsMissing")
class RequestParamsMissingError(ResponseError):
    """This class defines the exception for request parameters missing response error."""


@ResponseErrorRegister(404)
class NotFoundError(ResponseError):
    """This class defines the exception for 404 not found response error without error code."""


@ResponseErrorRegister(404, "ResourceNotExist")
class ResourceNotExistError(NotFoundError):
    """This class defines the exception for resource not existing response error."""


@ResponseErrorRegister(500, "InternalServerError")
class InternalServerError(ResponseError):
    """This class defines the exception for internal server error."""


@ResponseErrorRegister(401, "Unauthorized")
class UnauthorizedError(ResponseError):
    """This class defines the exception for unauthorized response error."""


@ResponseErrorRegister(503)
class ServiceUnavailableError(ResponseError):
    """This class defines the exception for 503 service unavailable error without error code."""



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the storage config."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do
from graviti.utility import SortParam


def list_storage_configs(
    access_key: str,
    url: str,
    workspace: str,
    *,
    sort: SortParam = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/workspaces/{workspace}/storage-configs`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.
        sort: The column and the direction the list result sorted by.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_storage_configs("ACCESSKEY-********", "https://api.graviti.com", "portex-test")
        {
            "default_storage_config": "AliCloud-oss-cn-shanghai",
            "storage_configs": [
                {
                    "id": "2bc95d506db2401b898067f1045d7f60",
                    "name": "AliCloud-oss-cn-shanghai",
                    "config_type": "GRAVITI",
                    "backend_type": "OSS"
                },
                {
                    "id": "2bc95d506db2401b898067f1045d7f61",
                    "name": "OSSConfig",
                    "config_type": "AUTHORIZED",
                    "backend_type": "OSS"
                }
            ],
            "offset": 0,
            "record_size": 2,
            "total_count": 2
        }

    """
    url = f"{url}/v2/workspaces/{workspace}/storage-configs"

    params = {
        "sort": sort,
        "offset": offset,
        "limit": limit,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def get_storage_config(
    access_key: str, url: str, workspace: str, storage_config: str
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/workspaces/{workspace}/storage-configs/{storage_config}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.
        storage_config: The name of the storage config.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_storage_config(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "portex-test",
        ...     "AliCloud-oss-cn-shanghai",
        ... )
        {
            "id": "2bc95d506db2401b898067f1045d7f60",
            "name": "AliCloud-oss-cn-shanghai",
            "config_type": "GRAVITI",
            "backend_type": "OSS",
        }

    """
    url = f"{url}/v2/workspaces/{workspace}/storage-configs/{storage_config}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def update_storage_configs(
    access_key: str, url: str, workspace: str, *, default_storage_config: str
) -> None:
    """Execute the OpenAPI `PATCH /v2/workspaces/{workspace}/storage-configs`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.
        default_storage_config: New default storage config.

    Examples:
        >>> update_storage_config(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "portex-test",
        ...     "AliCloud-oss-cn-shanghai",
        ... )

    """
    url = f"{url}/v2/workspaces/{workspace}/storage-configs"
    patch_data = {"deafault_storage_config": default_storage_config}

    open_api_do("PATCH", access_key, url, json=patch_data)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the workspace."""

from typing import Any, Dict

from graviti.openapi.requests import open_api_do


def get_current_workspace(access_key: str, url: str) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/current-workspace`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_current_workspace("ACCESSKEY-********", "https://api.graviti.com")
        {
            "type": "TEAM",
            "id": "41438e9df9a82a194e1e76cc31c1d8d4",
            "name": "portex-test",
            "alias": "PortexTest",
            "description": "xxx",
            "email": "xxxx@xx.com",
            "location": "xxx",
        }

    """
    url = f"{url}/v2/current-workspace"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def get_workspace(access_key: str, url: str, workspace: str) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/workspaces/{workspace}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_workspace("ACCESSKEY-********", "https://api.graviti.com", "portex-test")
        {
            "type": "TEAM",
            "id": "41438e9df9a82a194e1e76cc31c1d8d4",
            "name": "portex-test",
            "alias": "PortexTest",
            "description": "xxx",
            "email": "xxxx@xx.com",
            "location": "xxx",
        }

    """
    url = f"{url}/v2/workspaces/{workspace}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the actions."""

from typing import Any, Dict, Optional

from requests.models import Response

from graviti.openapi.requests import open_api_do
from graviti.utility import SortParam


def create_action(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    name: str,
    payload: str,
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{workspace}/{dataset}/actions`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        name: The name of the action.
        payload: The payload of the action.

    Returns:
        The response of OpenAPI.

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/actions"
    post_data = {"name": name, "payload": payload}
    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()


def list_actions(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    sort: SortParam = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/actions`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        sort: The column and the direction the list result sorted by.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/actions"

    params = {
        "sort": sort,
        "offset": offset,
        "limit": limit,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def get_action(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    action: str,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/actions/{action}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        action: The name of the aciton.

    Returns:
        The response of OpenAPI.

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/actions/{action}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def update_action(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    action: str,
    name: Optional[str] = None,
    state: Optional[str] = None,
    payload: Optional[str] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `PATCH /v2/datasets/{workspace}/{dataset}/actions/{action}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        action: The name of the aciton.
        name: The new name of the action.
        state: The new state of the action.
        payload: The new payload of the action.

    Returns:
        The response of OpenAPI.

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/actions/{action}"

    patch_data = {}
    if name is not None:
        patch_data["name"] = name
    if state is not None:
        patch_data["state"] = state
    if payload is not None:
        patch_data["payload"] = payload

    return open_api_do(  # type: ignore[no-any-return]
        "PATCH", access_key, url, json=patch_data
    ).json()


def delete_action(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    action: str,
) -> None:
    """Execute the OpenAPI `DELETE /v2/datasets/{workspace}/{dataset}/actions/{action}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        action: The name of the action.

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/actions/{action}"
    open_api_do("DELETE", access_key, url)


def create_action_run(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    action: str,
    arguments: Optional[Dict[str, Any]],
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{workspace}/{dataset}/actions/{action}/runs`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        action: The name of the action.
        arguments: The arguments of the action run.

    Returns:
        The response of OpenAPI.

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/actions/{action}/runs"

    post_data = {"arguments": arguments} if arguments is not None else None

    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()


def list_action_runs(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    action: str,
    sort: SortParam = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/actions/{action}/runs`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        action: The name of the action.
        sort: The column and the direction the list result sorted by.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/actions/{action}/runs"

    params = {
        "sort": sort,
        "offset": offset,
        "limit": limit,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def get_action_run(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    action: str,
    run_number: int,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/actions/{action}/runs\
    /{run_number}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        action: The name of the aciton.
        run_number: The number of the aciton run.

    Returns:
        The response of OpenAPI.

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/actions/{action}/runs/{run_number}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def cancel_action_run(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    action: str,
    run_number: int,
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{workspace}/{dataset}/actions/{action}/runs\
    /{run_number}/cancel`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        action: The name of the aciton.
        run_number: The number of the aciton run.

    Returns:
        The response of OpenAPI.

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/actions/{action}/runs/{run_number}/cancel"
    return open_api_do("POST", access_key, url).json()  # type: ignore[no-any-return]


def get_action_run_node_log(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    action: str,
    run_number: int,
    node_id: str,
) -> Response:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/actions/{action}/runs\
    /{run_number}/nodes/{node_id}/logs`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        action: The name of the aciton.
        run_number: The number of the aciton run.
        node_id: The id of the action run node.

    Returns:
        The response of OpenAPI.

    """
    url = (
        f"{url}/v2/datasets/{workspace}/{dataset}/actions/{action}/runs/{run_number}"
        f"/nodes/{node_id}/logs"
    )

    return open_api_do("GET", access_key, url)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the SearchHistory and SearchManager."""

from typing import TYPE_CHECKING, Any, Dict, Generator, Optional

from graviti.manager.commit import Commit
from graviti.manager.common import LIMIT
from graviti.manager.lazy import LazyPagingList
from graviti.openapi import (
    create_search_history,
    delete_search_history,
    get_search_history,
    list_search_histories,
)
from graviti.utility import ReprMixin, SortParam, check_type, convert_iso_to_datetime

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset


class SearchHistory(ReprMixin):  # pylint: disable=too-many-instance-attributes
    """This class defines the structure of the search history of Graviti Data Platform.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        response: The response of the OpenAPI associated with the search history::

                {
                    "id": <str>
                    "commit_id": <str>,
                    "draft_number": <int>,
                    "sheet": <str>,
                    "criteria": <dict>,
                    "total_count": <int>,
                    "creator": <str>,
                    "created_at": <str>,
                },

    Attributes:
        search_id: The id of this search history.
        commit: The commit of this search history.
        draft_number: The draft number of this search history.
        sheet: The sheet name of this search history.
        criteria: The criteria of this search history.
        total_count: The total count of this search history.
        creator: The creator of this search history.
        created_at: The create time of this search history.

    """

    _repr_attrs = (
        "commit",
        "draft_number",
        "sheet",
        "criteria",
        "total_count",
        "creator",
        "created_at",
    )

    def __init__(self, dataset: "Dataset", response: Dict[str, Any]) -> None:
        self._dataset = dataset
        self.search_id = response["id"]

        if "commit_id" in response:
            self.commit = Commit(dataset, response["commit_id"])
        else:
            self.draft_number = response["draft_number"]

        self.sheet = response["sheet"]
        self.criteria = response["criteria"]
        self.total_count = response["total_count"]
        self.creator = response["creator"]
        self.created_at = convert_iso_to_datetime(response["created_at"])

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.search_id}")'


class SearchManager:
    """This class defines the operations on the searches on Graviti.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.

    """

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _generate(
        self,
        offset: int,
        limit: int,
        *,
        commit_id: Optional[str],
        draft_number: Optional[int],
        sheet: Optional[str],
        sort: SortParam,
    ) -> Generator[SearchHistory, None, int]:
        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = list_search_histories(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            commit_id=commit_id,
            draft_number=draft_number,
            sheet=sheet,
            sort=sort,
            limit=limit,
            offset=offset,
        )

        for item in response["searches"]:
            yield SearchHistory(_dataset, item)

        return response["total_count"]  # type: ignore[no-any-return]

    def _create(
        self,
        commit_id: Optional[str],
        draft_number: Optional[int],
        sheet: str,
        criteria: Dict[str, Any],
    ) -> SearchHistory:
        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = create_search_history(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            commit_id=commit_id,
            draft_number=draft_number,
            sheet=sheet,
            criteria=criteria,
        )

        return SearchHistory(_dataset, response)

    def get(self, search_id: str) -> SearchHistory:
        """Get a Graviti search history with given search id.

        Arguments:
            search_id: The id of the search history.

        Returns:
            The requested :class:`~graviti.manager.search.SearchManager` instance.

        """
        check_type("search_id", search_id, str)

        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = get_search_history(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            search_id=search_id,
        )

        return SearchHistory(_dataset, response)

    def list(
        self,
        *,
        commit_id: Optional[str],
        draft_number: Optional[int],
        sheet: Optional[str],
        sort: SortParam,
    ) -> LazyPagingList[SearchHistory]:
        """List Graviti search histories.

        Arguments:
            commit_id: The commit id.
            draft_number: The draft number.
            sheet: The name of the sheet.
            sort: The column and the direction the list result sorted by.

        Returns:
            The LazyPagingList of :class:`~graviti.manager.search.SearchHistory` instances.

        """
        return LazyPagingList(
            lambda offset, limit: self._generate(
                offset,
                limit,
                commit_id=commit_id,
                draft_number=draft_number,
                sheet=sheet,
                sort=sort,
            ),
            LIMIT,
        )

    def delete(self, search_id: str) -> None:
        """Delete a Graviti search history with given search id.

        Arguments:
            search_id: The id of the search history.

        """
        check_type("search_id", search_id, str)

        _workspace = self._dataset.workspace
        delete_search_history(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self._dataset.name,
            search_id=search_id,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Paging list offset related class."""

from bisect import bisect_right
from itertools import accumulate, chain
from typing import Iterable, List, Tuple, TypeVar

_O = TypeVar("_O", bound="Offsets")


class Offsets:
    """The offsets manager of the paging list.

    Arguments:
        total_count: The total count of the elements in the paging list.
        limit: The size of each page.

    """

    _offsets: List[int]

    def __init__(self, total_count: int, limit: int) -> None:
        self.total_count = total_count
        self._limit = limit

    def _get_offsets(self) -> List[int]:
        if not hasattr(self, "_offsets"):
            self._offsets = (
                list(range(0, self.total_count, self._limit)) if self._limit != 0 else []
            )

        return self._offsets

    def update(self, start: int, stop: int, lengths: Iterable[int]) -> None:
        """Update the offsets when setting or deleting paging list items.

        Arguments:
            start: The start index.
            stop: The stop index.
            lengths: The length of the set values.

        """
        offsets = self._get_offsets()
        partial_offsets = list(accumulate(chain((offsets[start],), lengths)))
        try:
            last_offset = offsets[stop + 1]
        except IndexError:
            last_offset = self.total_count

        diff = partial_offsets.pop() - last_offset
        if diff != 0:
            self.total_count += diff
            for i in range(start + 1, len(offsets)):
                offsets[i] += diff

        offsets[start : stop + 1] = partial_offsets

    def get_coordinate(self, index: int) -> Tuple[int, int]:
        """Get the page coordinate of the elements.

        Arguments:
            index: The index of the element in paging list.

        Returns:
            The page number and the index of the page.

        """
        if not hasattr(self, "_offsets"):
            try:
                return divmod(index, self._limit)
            except ZeroDivisionError:
                return 0, index

        i = bisect_right(self._offsets, index) - 1
        return i, index - self._offsets[i]

    def extend(self, lengths: Iterable[int]) -> None:
        """Update the offsets when extending the paging list.

        Arguments:
            lengths: The lengths of the extended pages.

        """
        offsets = self._get_offsets()
        offsets.extend(accumulate(chain((self.total_count,), lengths)))
        self.total_count = offsets.pop()

    def copy(self: _O) -> _O:
        """Return a copy of the Offsets.

        Returns:
            A copy of the Offsets.

        """
        obj = self.__class__(self.total_count, self._limit)
        if hasattr(self, "_offsets"):
            obj._offsets = self._offsets.copy()  # pylint: disable=protected-access

        return obj



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""OpenAPI module."""

from graviti.openapi.action import (
    cancel_action_run,
    create_action,
    create_action_run,
    delete_action,
    get_action,
    get_action_run,
    get_action_run_node_log,
    list_action_runs,
    list_actions,
    update_action,
)
from graviti.openapi.branch import create_branch, delete_branch, get_branch, list_branches
from graviti.openapi.commit import commit_draft, get_commit, get_revision, list_commits
from graviti.openapi.data import (
    add_data,
    delete_data,
    list_commit_data,
    list_draft_data,
    update_data,
)
from graviti.openapi.dataset import (
    create_dataset,
    delete_dataset,
    get_dataset,
    list_datasets,
    update_dataset,
)
from graviti.openapi.draft import create_draft, get_draft, list_drafts, update_draft
from graviti.openapi.object import copy_objects, get_object_permission
from graviti.openapi.requests import do
from graviti.openapi.schema import update_schema
from graviti.openapi.search import (
    create_search,
    create_search_history,
    delete_search_history,
    get_search_history,
    get_search_total_count,
    list_search_histories,
    list_search_records,
)
from graviti.openapi.sheet import (
    create_sheet,
    delete_sheet,
    get_commit_sheet,
    get_draft_sheet,
    list_commit_sheets,
    list_draft_sheets,
)
from graviti.openapi.storage_config import (
    get_storage_config,
    list_storage_configs,
    update_storage_configs,
)
from graviti.openapi.tag import create_tag, delete_tag, get_tag, list_tags
from graviti.openapi.user import get_current_user
from graviti.openapi.workspace import get_current_workspace, get_workspace

RECORD_KEY = "__record_key"

__all__ = [
    "RECORD_KEY",
    "add_data",
    "cancel_action_run",
    "commit_draft",
    "copy_objects",
    "create_action",
    "create_action_run",
    "create_branch",
    "create_dataset",
    "create_draft",
    "create_search",
    "create_search_history",
    "create_sheet",
    "create_tag",
    "delete_action",
    "delete_branch",
    "delete_data",
    "delete_dataset",
    "delete_search_history",
    "delete_sheet",
    "delete_tag",
    "do",
    "get_action",
    "get_action_run",
    "get_action_run_node_log",
    "get_branch",
    "get_commit",
    "get_commit_sheet",
    "get_current_user",
    "get_current_workspace",
    "get_dataset",
    "get_draft",
    "get_draft_sheet",
    "get_object_permission",
    "get_revision",
    "get_search_history",
    "get_search_total_count",
    "get_storage_config",
    "get_tag",
    "get_workspace",
    "list_action_runs",
    "list_actions",
    "list_branches",
    "list_commit_data",
    "list_commit_sheets",
    "list_commits",
    "list_datasets",
    "list_draft_data",
    "list_draft_sheets",
    "list_drafts",
    "list_search_histories",
    "list_search_records",
    "list_storage_configs",
    "list_tags",
    "update_action",
    "update_data",
    "update_dataset",
    "update_draft",
    "update_schema",
    "update_storage_configs",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Workspace."""

from graviti.manager.dataset import DatasetManager
from graviti.manager.storage_config import StorageConfigManager
from graviti.openapi import get_current_workspace
from graviti.utility import urlnorm


class Workspace:  # pylint: disable=too-many-instance-attributes
    """This class defines the initial client to interact between local and server.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.

    """

    _DEFAULT_URL_CN = "https://api.graviti.cn"
    _DEFAULT_URL_COM = "https://api.graviti.com"

    def __init__(
        self,
        access_key: str,
        url: str = "",
    ) -> None:
        if access_key.startswith("Accesskey-"):
            url = url if url else Workspace._DEFAULT_URL_CN
        elif access_key.startswith("ACCESSKEY-"):
            url = url if url else Workspace._DEFAULT_URL_COM
        else:
            raise TypeError("Wrong accesskey format!")

        if not url.startswith("https://"):
            raise TypeError("Invalid url, only support url starts with 'https://'")

        url = urlnorm(url)

        self.access_key = access_key
        self.url = url

        response = get_current_workspace(access_key, url)

        self._id = response["id"]
        self.type = response["type"]

        self.name = response["name"]
        self.alias = response["alias"]
        self.description = response["description"]
        self.email = response["email"]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'

    @property
    def datasets(self) -> DatasetManager:
        """Get class :class:`~graviti.manager.dataset.DatasetManager` instance.

        Returns:
            Required :class:`~graviti.manager.dataset.DatasetManager` instance.

        """
        return DatasetManager(self)

    @property
    def storage_configs(self) -> StorageConfigManager:
        """Get class :class:`~graviti.manager.storage_config.StorageConfigManager` instance.

        Returns:
            Required :class:`~graviti.manager.storage_config.StorageConfigManager` instance.

        """
        return StorageConfigManager(self)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""The table-structured data container related classes."""


from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Tuple, Type, TypeVar, Union

import pyarrow as pa

from graviti.paging import LazyFactoryBase
from graviti.portex import PortexType

if TYPE_CHECKING:
    import pandas

    from graviti.dataframe.frame import DataFrame
    from graviti.manager.permission import ObjectPermissionManager

_T = TypeVar("_T", bound="Container")


class Container:
    """The base class for the table-structured data container."""

    schema: PortexType
    _root: Optional["DataFrame"] = None
    _name: Tuple[str, ...] = ()

    def __len__(self) -> int:
        raise NotImplementedError

    @classmethod
    def _create(
        cls: Type[_T],
        schema: PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _T:
        obj: _T = object.__new__(cls)
        obj.schema = schema
        obj._root = root
        obj._name = name

        return obj

    @classmethod
    def _from_factory(
        cls: Type[_T],
        factory: LazyFactoryBase,
        schema: PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
        *,
        object_permission_manager: Optional["ObjectPermissionManager"] = None,
    ) -> _T:
        raise NotImplementedError

    @classmethod
    def _from_pyarrow(
        cls: Type[_T],
        array: pa.Array,
        schema: PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
        *,
        object_permission_manager: Optional["ObjectPermissionManager"] = None,
    ) -> _T:
        raise NotImplementedError

    @classmethod
    def _from_iterable(
        cls: Type[_T],
        array: Iterable[Any],
        schema: PortexType,
    ) -> _T:
        raise NotImplementedError

    def _extend(self: _T, values: _T) -> None:
        raise NotImplementedError

    def _repr_folding(self) -> str:
        raise NotImplementedError

    def _copy(
        self: _T, schema: PortexType, root: Optional["DataFrame"] = None, name: Tuple[str, ...] = ()
    ) -> _T:
        raise NotImplementedError

    def _get_item_by_location(self, key: int) -> Any:
        raise NotImplementedError

    def _get_slice_by_location(
        self: _T,
        key: slice,
        schema: PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _T:
        raise NotImplementedError

    def _del_item_by_location(self, key: Union[int, slice]) -> None:
        raise NotImplementedError

    def _refresh_data_from_factory(
        self,
        factory: LazyFactoryBase,
        object_permission_manager: Optional["ObjectPermissionManager"],
    ) -> None:
        raise NotImplementedError

    def _to_pandas_series(self) -> "pandas.Series":
        raise NotImplementedError

    # TODO: Defines a base indexer for the loc and iloc return type.
    @property
    def iloc(self) -> Any:
        """Purely integer-location based indexing for selection by position.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    @property
    def loc(self) -> Any:
        """Access the row by index.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def to_pylist(self, *, _to_backend: bool = False) -> List[Any]:
        """Convert the container to a python list.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def to_pandas(self) -> Union["pandas.Series", "pandas.DataFrame"]:
        """Convert the graviti Container to a pandas Series or DataFrame.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def copy(self: _T) -> _T:
        """Get a copy of the container.

        Returns:
            A copy of the container.

        """
        return self._copy(self.schema.copy())



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Commit and CommitManager."""

from datetime import datetime
from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, Type, TypeVar, Union

from graviti.dataframe import DataFrame
from graviti.exception import NoCommitsError
from graviti.manager.common import CURRENT_COMMIT, LIMIT, check_head_status
from graviti.manager.lazy import LazyPagingList
from graviti.manager.sheets import Sheets
from graviti.openapi import (
    create_search,
    get_commit,
    get_commit_sheet,
    get_revision,
    list_commit_data,
    list_commit_sheets,
    list_commits,
)
from graviti.operation import SheetOperation
from graviti.paging import LazyLowerCaseFactory
from graviti.portex import PortexRecordBase
from graviti.utility import LazyAttr, check_type, convert_iso_to_datetime

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset

_C = TypeVar("_C", bound="Commit")
_NC = TypeVar("_NC", bound="NamedCommit")


class Commit(Sheets):  # pylint: disable=too-many-instance-attributes
    """This class defines the structure of a commit.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        commit_id: The commit id.

    Attributes:
        commit_id: The commit id of the commit.
        parent: The parent commit of the commit.
        title: The title of the commit.
        description: The description of the commit.
        committer: The committer of the commit.
        committed_at: The commit time of the commit.

    """

    _repr_attrs: Tuple[str, ...] = ("parent", "title", "committer", "committed_at")

    parent = LazyAttr[Optional["Commit"]]()
    title = LazyAttr[str]()
    description = LazyAttr[str]()
    committer = LazyAttr[str]()
    committed_at = LazyAttr[datetime]()

    def __init__(
        self,
        dataset: "Dataset",
        commit_id: Optional[str],
    ) -> None:
        self._dataset = dataset
        self.commit_id = commit_id
        self.operations: List[SheetOperation] = []

    def _init(self, response: Dict[str, Any]) -> None:
        parent_commit_id = response["parent_commit_id"]
        self.parent = None if parent_commit_id is None else Commit(self._dataset, parent_commit_id)
        self.title = response["title"]
        self.description = response["description"]
        self.committer = response["committer"]
        self.committed_at = convert_iso_to_datetime(response["committed_at"])

    def _load(self) -> None:
        if self.commit_id is None:
            raise AttributeError(
                "The attribute is not available due to this commit has no commit history."
            )

        _workspace = self._dataset.workspace
        response = get_commit(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self._dataset.name,
            commit_id=self.commit_id,
        )
        self._init(response)

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.commit_id}")'

    def _list_data(self, offset: int, limit: int, sheet_name: str) -> Dict[str, Any]:
        _workspace = self._dataset.workspace
        return list_commit_data(  # type: ignore[no-any-return]
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self._dataset.name,
            commit_id=self.commit_id,  # type: ignore[arg-type]
            sheet=sheet_name,
            offset=offset,
            limit=limit,
        )["data"]

    def _list_sheets(self) -> Dict[str, Any]:
        if self.commit_id is None:
            return {"sheets": []}

        _workspace = self._dataset.workspace
        return list_commit_sheets(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self._dataset.name,
            commit_id=self.commit_id,
        )

    def _get_sheet(self, sheet_name: str) -> Dict[str, Any]:
        _workspace = self._dataset.workspace
        return get_commit_sheet(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self._dataset.name,
            commit_id=self.commit_id,  # type: ignore[arg-type]
            sheet=sheet_name,
            with_record_count=True,
        )

    def _init_dataframe(self, sheet_name: str) -> DataFrame:
        df = super()._init_dataframe(sheet_name)
        df.searcher = partial(self.search, sheet_name)
        return df

    @classmethod
    def from_response(cls: Type[_C], dataset: "Dataset", response: Dict[str, Any]) -> _C:
        """Create a :class:`Commit` instance from python dict.

        Arguments:
            dataset: The dataset of the commit.
            response: A python dict containing all the information of the commit::

                {
                    "commit_id": <str>
                    "parent_commit_id": <Optional[str]>
                    "title": <str>
                    "description": <str>
                    "committer":  <str>
                    "committed_at": <str>
                }

        Returns:
            A :class:`Commit` instance created from the input python dict.

        """
        obj = cls(dataset, response["commit_id"])
        obj._init(response)

        return obj

    def search(
        self, sheet: str, criteria: Dict[str, Any], schema: Optional[PortexRecordBase] = None
    ) -> DataFrame:
        """Create a search.

        Arguments:
            sheet: The sheet name.
            criteria: The criteria of search.
            schema: The schema of the search result DataFrame.

        Raises:
            NoCommitsError: When there is no commit on the current branch.

        Returns:
            The created :class:`~graviti.dataframe.DataFrame` instance.

        """
        if self.commit_id is None:
            raise NoCommitsError("No commit on the current branch. Please commit a draft first")

        _dataset = self._dataset
        _workspace = _dataset.workspace
        search_id = create_search(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            commit_id=self.commit_id,
            sheet=sheet,
            criteria=criteria,
            offset=0,
            limit=1,
        )["search_id"]

        def _getter(
            offset: Optional[int] = None,
            limit: Optional[int] = None,
            getter_criteria: Dict[str, Any] = criteria,
        ) -> List[Dict[str, Any]]:
            return create_search(  # type: ignore[no-any-return]
                _workspace.access_key,
                _workspace.url,
                _workspace.name,
                _dataset.name,
                commit_id=self.commit_id,  # type: ignore[arg-type]
                sheet=sheet,
                search_id=search_id,
                criteria=getter_criteria,
                offset=offset,
                limit=limit,
            )["data"]

        count_criteria = criteria.copy()
        count_criteria["select"] = [{"count": {"$count": ["$."]}}]
        total_count = _getter(getter_criteria=count_criteria)[0]["count"]

        if schema is None:
            schema = PortexRecordBase.from_yaml(self._get_sheet(sheet)["schema"])

        factory = LazyLowerCaseFactory(
            total_count,
            LIMIT,
            _getter,
            schema.to_pyarrow(_to_backend=True),
        )

        return DataFrame._from_factory(  # pylint: disable=protected-access
            factory, schema, object_permission_manager=_dataset.object_permission_manager
        )


class NamedCommit(Commit):
    """This class defines the structure of a named commit.

    :class:`NamedCommit` is the base class of :class:`~graviti.manager.branch.Branch`
    and :class:`~graviti.manager.tag.Tag`.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        name: The name of the named commit.
        commit_id: The commit id.

    """

    _repr_attrs = ("commit_id",) + Commit._repr_attrs

    def __init__(
        self,
        dataset: "Dataset",
        name: str,
        commit_id: Optional[str],
    ) -> None:
        super().__init__(dataset, commit_id)
        self.name = name

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'

    @classmethod
    def from_response(cls: Type[_NC], dataset: "Dataset", response: Dict[str, Any]) -> _NC:
        """Create a :class:`NamedCommit` instance from python dict.

        Arguments:
            dataset: The dataset of the NamedCommit.
            response: A python dict containing all the information of the NamedCommit::

                {
                    "name": <str>
                    "commit_id": <Optional[str]>
                    "parent_commit_id": <Optional[str]>
                    "title": <str>
                    "description": <str>
                    "committer":  <str>
                    "committed_at": <str>
                }

        Returns:
            A :class:`NamedCommit` instance created from the input python dict.

        """
        obj = cls(dataset, response["name"], response["commit_id"])
        obj._init(response)

        return obj


class CommitManager:
    """This class defines the operations on the commit in Graviti.

    Arguments:
        dataset: :class:`~graviti.manager.dataset.Dataset` instance.

    """

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _generate(self, revision: str, offset: int, limit: int) -> Generator[Commit, None, int]:
        _dataset = self._dataset
        _workspace = _dataset.workspace
        head = _dataset.HEAD

        response = list_commits(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            revision=revision,
            offset=offset,
            limit=limit,
        )

        commits = response["commits"]
        if offset == 0 and commits:
            check_head_status(head, revision, commits[0]["commit_id"])

        for item in commits:
            yield Commit.from_response(_dataset, item)

        return response["total_count"]  # type: ignore[no-any-return]

    def get(self, revision: str = CURRENT_COMMIT) -> Commit:
        """Get the certain commit with the given revision.

        Arguments:
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name. The default value is the current commit of the
                dataset.

        Raises:
            NoCommitsError: When revision is not given and the commit id of current dataset is None,
                or when the given branch has no commit history yet.

        Returns:
            The :class:`.Commit` instance with the given revision.

        """
        _dataset = self._dataset
        _workspace = _dataset.workspace
        head = _dataset.HEAD
        if revision is CURRENT_COMMIT:
            _revision = head.commit_id
            if _revision is None:
                raise NoCommitsError("No commits on the default branch yet")
        else:
            check_type("revision", revision, str)
            _revision = revision

        response = get_revision(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            revision=_revision,
        )
        if response["commit_id"] is None:
            raise NoCommitsError("No commits on the default branch yet")

        check_head_status(head, _revision, response["commit_id"])

        return Commit.from_response(_dataset, response)

    def list(
        self, revision: str = CURRENT_COMMIT
    ) -> Union[LazyPagingList[Commit], List[Optional[Commit]]]:
        """List the commits.

        Arguments:
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name.
                If it is given, list the commits before the given commit.
                If it is not given, list the commits before the current commit.

        Returns:
            The LazyPagingList of :class:`commits<.Commit>` instances.

        """
        if revision is CURRENT_COMMIT:
            _revision = self._dataset.HEAD.commit_id
            if _revision is None:
                return []
        else:
            check_type("revision", revision, str)
            _revision = revision

        return LazyPagingList(
            lambda offset, limit: self._generate(
                _revision, offset, limit  # type: ignore[arg-type]
            ),
            LIMIT,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the sheet."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do


def _list_sheet(
    access_key: str,
    url: str,
    with_record_count: Optional[bool],
    offset: Optional[int],
    limit: Optional[int],
) -> Dict[str, Any]:

    params = {
        "with_record_count": with_record_count,
        "offset": offset,
        "limit": limit,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def _get_sheet(
    access_key: str,
    url: str,
    with_record_count: Optional[bool],
    schema_format: Optional[str],
) -> Dict[str, Any]:
    params = {
        "with_record_count": with_record_count,
        "schema_format": schema_format,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def create_sheet(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    draft_number: int,
    name: str,
    schema: str,
    _avro_schema: str,
    _arrow_schema: Optional[str] = None,
    record_key_strategy: Optional[str] = None,
) -> None:
    """Execute the OpenAPI `POST /v2/datasets/{workspace}/{dataset}/drafts/{draft_number}/sheets`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        name: The sheet name.
        schema: The portex schema of the sheet.
        record_key_strategy: The ``__record_key`` generation strategy.
            If None, it is batch auto-increment sorting record key.

    Examples:
        >>> create_sheet(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     draft_number = 1,
        ...     name = "val",
        ...     schema = '{"imports": [{"repo": "https://github.com/Project-OpenBytes/portex-\
standard@main", "types": [{"name": "file.Image"}]}], "type": "record", "fields": [{"name": \
"filename", "type": "string"}, {"name": "image", "type": "file.Image"}]}',
        ...     _avro_schema = '{"type": "record", "name": "root", "namespace": "cn.graviti.portex"\
, "aliases": [], "fields": [{"name": "filename", "type": "string"}, {"name": "image", "type": \
{"type": "record", "name": "image", "namespace": "cn.graviti.portex.root", "aliases": [], \
"fields": [{"name": "checksum", "type": [null, "string"]}]}}]}',
        ... )

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/drafts/{draft_number}/sheets"
    post_data = {"name": name, "schema": schema, "_avro_schema": _avro_schema}

    if _arrow_schema is not None:
        post_data["_arrow_schema"] = _arrow_schema
    if record_key_strategy is not None:
        post_data["record_key_strategy"] = record_key_strategy

    open_api_do("POST", access_key, url, json=post_data)


def list_draft_sheets(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    draft_number: int,
    with_record_count: Optional[bool] = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/drafts/{draft_number}/sheets`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        with_record_count: Whether return the record count of each sheet. The default value of
            this param in OpenAPI is False.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_draft_sheets(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     draft_number = 1,
        ... )
        {
            "sheets": [
                {
                    "name": "test",
                    "created_at": "2021-03-03T18:58:10Z",
                    "updated_at": "2021-03-04T18:58:10Z",
                },
                {
                    "name": "trainval",
                    "created_at": "2021-03-05T18:58:10Z",
                    "updated_at": "2021-03-06T18:58:10Z",
                }
            ],
            "offset": 0,
            "record_size": 2,
            "total_count": 2
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/drafts/{draft_number}/sheets"

    return _list_sheet(
        access_key, url, with_record_count=with_record_count, offset=offset, limit=limit
    )


def list_commit_sheets(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    commit_id: str,
    with_record_count: Optional[bool] = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/commits/{commit_id}/sheets`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        commit_id: The commit id.
        with_record_count: Whether return the record count of each sheet. The default value of
            this param in OpenAPI is False.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_commit_sheets(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     commit_id = "fde63f357daf46088639e9f57fd81cad",
        ... )
        {
            "sheets": [
                {
                    "name": "test",
                    "created_at": "2021-03-03T18:58:10Z",
                    "updated_at": "2021-03-04T18:58:10Z",
                },
                {
                    "name": "trainval",
                    "created_at": "2021-03-05T18:58:10Z",
                    "updated_at": "2021-03-06T18:58:10Z",
                }
            ],
            "offset": 0,
            "record_size": 2,
            "total_count": 2
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/commits/{commit_id}/sheets"

    return _list_sheet(
        access_key, url, with_record_count=with_record_count, offset=offset, limit=limit
    )


def get_draft_sheet(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
    with_record_count: Optional[bool] = None,
    schema_format: Optional[str] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/drafts/{draft_number}/sheets`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        sheet: The sheet name.
        with_record_count: Whether return the record count of each sheet. The default value of
            this param in OpenAPI is False.
        schema_format: Fill "JSON"/"YAML" to determine whether the schema_format of the returned
            schema is json or yaml. None means "JSON" format.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_draft_sheet(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     draft_number = 1,
        ...     sheet = "sheet-2",
        ...     with_record_count=True,
        ... )
        {
            "name": "trainval",
            "created_at": "2021-03-05T18:58:10Z",
            "updated_at": "2021-03-06T18:58:10Z",
            "record_count": 10000,
            "schema": '{"imports": [{"repo": "https://github.com/Project-OpenBytes/...'
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/drafts/{draft_number}/sheets/{sheet}"

    return _get_sheet(
        access_key, url, with_record_count=with_record_count, schema_format=schema_format
    )


def get_commit_sheet(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    commit_id: str,
    sheet: str,
    with_record_count: Optional[bool] = None,
    schema_format: Optional[str] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/commits/\
    {commit_id}/sheets/{sheet}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        commit_id: The commit id..
        sheet: The sheet name.
        with_record_count: Whether return the record count of each sheet. The default value of
            this param in OpenAPI is False.
        schema_format: Fill "JSON"/"YAML" to determine whether the schema_format of the returned
            schema is json or yaml. None means "JSON" format.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_commit_sheet(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     commit_id = "fde63f357daf46088639e9f57fd81cad",
        ...     sheet = "sheet-2",
        ...     with_record_count=True,
        ... )
        {
            "name": "trainval",
            "created_at": "2021-03-05T18:58:10Z",
            "updated_at": "2021-03-06T18:58:10Z",
            "record_count": 10000,
            "schema": '{"imports": [{"repo": "https://github.com/Project-OpenBytes/...'
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/commits/{commit_id}/sheets/{sheet}"

    return _get_sheet(
        access_key, url, with_record_count=with_record_count, schema_format=schema_format
    )


def delete_sheet(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
) -> None:
    """Execute the OpenAPI `DELETE /v2/datasets/{workspace}/{dataset}/drafts/\
    {draft_number}/sheets/{sheet}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        sheet: The name of the sheet to be deleted.

    Examples:
        >>> delete_sheet(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     draft_number=1,
        ...     sheet="sheet-2"
        ... )

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/drafts/{draft_number}/sheets/{sheet}"
    open_api_do("DELETE", access_key, url)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti indexing related class."""

from typing import TYPE_CHECKING, Any, Iterable, Union, overload

if TYPE_CHECKING:
    from graviti.dataframe.row.series import Series as RowSeries


class RowSeriesILocIndexer:
    """Index class for RowSeries.iloc."""

    def __init__(self, obj: "RowSeries") -> None:
        self.obj: "RowSeries" = obj

    # @overload
    # def __getitem__(self, key: Union[Iterable[bool], slice]) -> "RowSeries":
    #    ...

    @overload
    def __getitem__(self, key: int) -> Any:
        ...

    @overload
    def __getitem__(self, key: Iterable[int]) -> "RowSeries":
        ...

    def __getitem__(self, key: Union[int, Iterable[int]]) -> Any:
        return self.obj._getitem_by_location(key)

    def __setitem__(self, key: int, value: Any) -> None:
        pass


class RowSeriesLocIndexer:
    """Index class for RowSeries.loc."""

    def __init__(self, obj: "RowSeries") -> None:
        self.obj: "RowSeries" = obj

    # @overload
    # def __getitem__(self, key: Union[Iterable[bool], slice]) -> "RowSeries":
    #    ...

    @overload
    def __getitem__(self, key: str) -> Any:
        ...

    @overload
    def __getitem__(self, key: Iterable[str]) -> "RowSeries":
        ...

    def __getitem__(self, key: Union[str, Iterable[str]]) -> Any:
        return self.obj[key]

    def __setitem__(self, key: str, value: Any) -> None:
        pass
